@@ -3,6 +3,9 @@
 
 #include <stan/io/var_context.hpp>
 #include <stan/math/rev/core.hpp>
+#ifdef STAN_MODEL_FVAR_VAR
+#include <stan/math/mix.hpp>
+#endif
 #include <stan/model/prob_grad.hpp>
 #include <boost/random/additive_combine.hpp>
 #include <ostream>
@@ -53,6 +56,14 @@ class model_base : public prob_grad {
    */
   virtual std::string model_name() const = 0;
 
+  /**
+   * Returns the compile information of the model:
+   * stanc version and stanc flags used to compile the model.
+   *
+   * @return model name
+   */
+  virtual std::vector<std::string> model_compile_info() const = 0;
+
   /**
    * Set the specified argument to sequence of parameters, transformed
    * parameters, and generated quantities in the order in which they
@@ -60,10 +71,20 @@ class model_base : public prob_grad {
    *
    * @param[in,out] names sequence of names parameters, transformed
    * parameters, and generated quantities
+   * @param[in] include_tparams true if transformed parameters should
+   * be included
+   * @param[in] include_gqs true if generated quantities should be
+   * included
    */
-  virtual void get_param_names(std::vector<std::string>& names) const = 0;
-
+  virtual void get_param_names(std::vector<std::string>& names,
+                               bool include_tparams = true,
+                               bool include_gqs = true) const = 0;
   /**
+   * WARNING: This function bakes in the assumption that the
+   * parameter values are rectangular. This is already not true
+   * for Tuple types, and will be fundamentally broken for ragged
+   * arrays or sparse matrices.
+   *
    * Set the dimensionalities of constrained parameters, transformed
    * parameters, and generated quantities.  The input sequence is
    * cleared and resized.  The dimensions of each parameter
@@ -79,15 +100,20 @@ class model_base : public prob_grad {
    * dimensionality `std::vector<size_t>{2, 3, 4}`.
    *
    * @param[in,out] dimss sequence of dimensions specifications to set
+   * @param[in] include_tparams true if transformed parameters should
+   * be included
+   * @param[in] include_gqs true if generated quantities should be
+   * included
    */
-  virtual void get_dims(std::vector<std::vector<size_t> >& dimss) const = 0;
-
+  virtual void get_dims(std::vector<std::vector<size_t> >& dimss,
+                        bool include_tparams = true,
+                        bool include_gqs = true) const = 0;
   /**
    *  Set the specified sequence to the indexed, scalar, constrained
    *  parameter names.  Each variable is output with a
    *  period-separated list of indexes as a suffix, indexing from 1.
    *
-   * <p>A real parmeter `alpha` will produce output `alpha` with no
+   * <p>A real parameter `alpha` will produce output `alpha` with no
    * indexes.
    *
    * <p>A 3-dimensional vector (row vector, simplex, etc.)
@@ -100,7 +126,7 @@ class model_base : public prob_grad {
    * X.2.3`.
    *
    * <p>Arrays are handled in natural C++ order, 2 x 3 x 4 array `a`
-   * will produce ouptut `a.1.1.1`, `a.1.1.2`, `a.1.1.3`, `a.1.1.4`,
+   * will produce output `a.1.1.1`, `a.1.1.2`, `a.1.1.3`, `a.1.1.4`,
    * `a.1.2.1`, `a.1.2.2`, `a.1.2.3`, `a.1.2.4`, `a.1.3.1`, `a.1.3.2`,
    * `a.1.3.3`, `a.1.3.4`, `a.2.1.1`, `a.2.1.2`, `a.2.1.3`, `a.2.1.4`,
    * `a.2.2.1`, `a.2.2.2`, `a.2.2.3`, `a.2.2.4`, `a.2.3.1`, `a.2.3.2`,
@@ -186,7 +212,7 @@ class model_base : public prob_grad {
    * parameters to constrained parameters; full details for Stan
    * language types can be found in the language reference manual.
    *
-   * @param[in] params_r unconstrained parmeters
+   * @param[in] params_r unconstrained parameters
    * @param[in,out] msgs message stream
    * @return log density for specified parameters
    */
@@ -347,6 +373,21 @@ class model_base : public prob_grad {
                            bool include_tparams = true, bool include_gqs = true,
                            std::ostream* msgs = 0) const = 0;
 
+  /**
+   * Convert the specified sequence of constrained parameters to a
+   * sequence of unconstrained parameters.
+   *
+   * This is the inverse of write_array. The output will be resized
+   * if necessary to match the number of unconstrained parameters.
+   *
+   * @param[in] params_r_constrained constrained parameters input
+   * @param[in,out] params_r unconstrained parameters produced
+   * @param[in,out] msgs msgs stream to which messages are written
+   */
+  virtual void unconstrain_array(const Eigen::VectorXd& params_r_constrained,
+                                 Eigen::VectorXd& params_r,
+                                 std::ostream* msgs = nullptr) const = 0;
+
   // TODO(carpenter): cut redundant std::vector versions from here ===
 
   /**
@@ -583,6 +624,84 @@ class model_base : public prob_grad {
                            std::vector<double>& params_r_constrained,
                            bool include_tparams = true, bool include_gqs = true,
                            std::ostream* msgs = 0) const = 0;
+
+  /**
+   * Convert the specified sequence of constrained parameters to a
+   * sequence of unconstrained parameters.
+   *
+   * This is the inverse of write_array. The output will be resized
+   * if necessary to match the number of unconstrained parameters.
+   *
+   * @param[in] params_r_constrained constrained parameters input
+   * @param[in,out] params_r unconstrained parameters produced
+   * @param[in,out] msgs msgs stream to which messages are written
+   */
+  virtual void unconstrain_array(
+      const std::vector<double>& params_r_constrained,
+      std::vector<double>& params_r, std::ostream* msgs = nullptr) const = 0;
+
+#ifdef STAN_MODEL_FVAR_VAR
+
+  /**
+   * Return the log density for the specified unconstrained
+   * parameters, without Jacobian and with normalizing constants for
+   * probability functions.
+   *
+   * @param[in] params_r unconstrained parameters
+   * @param[in,out] msgs message stream
+   * @return log density for specified parameters
+   */
+  virtual math::fvar<math::var> log_prob(
+      Eigen::Matrix<math::fvar<math::var>, -1, 1>& params_r,
+      std::ostream* msgs) const = 0;
+
+  /**
+   * Return the log density for the specified unconstrained
+   * parameters, with Jacobian correction for constraints and with
+   * normalizing constants for probability functions.
+   *
+   * <p>The Jacobian is of the inverse transform from unconstrained
+   * parameters to constrained parameters; full details for Stan
+   * language types can be found in the language reference manual.
+   *
+   * @param[in] params_r unconstrained parameters
+   * @param[in,out] msgs message stream
+   * @return log density for specified parameters
+   */
+  virtual math::fvar<math::var> log_prob_jacobian(
+      Eigen::Matrix<math::fvar<math::var>, -1, 1>& params_r,
+      std::ostream* msgs) const = 0;
+
+  /**
+   * Return the log density for the specified unconstrained
+   * parameters, without Jacobian correction for constraints and
+   * dropping normalizing constants.
+   *
+   * @param[in] params_r unconstrained parameters
+   * @param[in,out] msgs message stream
+   * @return log density for specified parameters
+   */
+  virtual math::fvar<math::var> log_prob_propto(
+      Eigen::Matrix<math::fvar<math::var>, -1, 1>& params_r,
+      std::ostream* msgs) const = 0;
+
+  /**
+   * Return the log density for the specified unconstrained
+   * parameters, with Jacobian correction for constraints and dropping
+   * normalizing constants.
+   *
+   * <p>The Jacobian is of the inverse transform from unconstrained
+   * parameters to constrained parameters; full details for Stan
+   * language types can be found in the language reference manual.
+   *
+   * @param[in] params_r unconstrained parameters
+   * @param[in,out] msgs message stream
+   * @return log density for specified parameters
+   */
+  virtual math::fvar<math::var> log_prob_propto_jacobian(
+      Eigen::Matrix<math::fvar<math::var>, -1, 1>& params_r,
+      std::ostream* msgs) const = 0;
+#endif
 };
 
 }  // namespace model
