@@ -16,14 +16,11 @@ namespace util {
  * @return var_context
  */
 inline stan::io::dump create_unit_e_diag_inv_metric(size_t num_params) {
+  std::string dims("),.Dim=c(" + std::to_string(num_params) + "))");
+  Eigen::IOFormat RFmt(Eigen::StreamPrecision, Eigen::DontAlignCols, ", ", ",",
+                       "", "", "inv_metric <- structure(c(", dims);
   std::stringstream txt;
-  txt << "inv_metric <- structure(c(";
-  for (size_t i = 0; i < num_params; ++i) {
-    txt << "1.0";
-    if (i < num_params - 1)
-      txt << ", ";
-  }
-  txt << "),.Dim=c(" << num_params << "))";
+  txt << Eigen::VectorXd::Ones(num_params).format(RFmt);
   return stan::io::dump(txt);
 }
 }  // namespace util
