@@ -3,7 +3,7 @@
  *                David J. Gardner, Radu Serban and Aaron Collier @ LLNL
  *-----------------------------------------------------------------
  * SUNDIALS Copyright Start
- * Copyright (c) 2002-2019, Lawrence Livermore National Security
+ * Copyright (c) 2002-2022, Lawrence Livermore National Security
  * and Southern Methodist University.
  * All rights reserved.
  *
@@ -44,9 +44,11 @@
   ---------------------------------------------------------------*/
 int KINSetLinearSolver(void *kinmem, SUNLinearSolver LS, SUNMatrix A)
 {
-  KINMem   kin_mem;
-  KINLsMem kinls_mem;
-  int      retval, LSType;
+  KINMem      kin_mem;
+  KINLsMem    kinls_mem;
+  int         retval, LSType;
+  booleantype iterative;    /* is the solver iterative?    */
+  booleantype matrixbased;  /* is a matrix structure used? */
 
   /* Return immediately if either kinmem or LS inputs are NULL */
   if (kinmem == NULL) {
@@ -63,16 +65,28 @@ int KINSetLinearSolver(void *kinmem, SUNLinearSolver LS, SUNMatrix A)
   kin_mem = (KINMem) kinmem;
 
   /* Test if solver is compatible with LS interface */
-  if ( (LS->ops->gettype == NULL) ||
-       (LS->ops->initialize == NULL) ||
-       (LS->ops->setup == NULL) ||
-       (LS->ops->solve == NULL) ) {
+  if ( (LS->ops->gettype == NULL) || (LS->ops->solve == NULL) ) {
     KINProcessError(kin_mem, KINLS_ILL_INPUT, "KINLS",
                    "KINSetLinearSolver",
                    "LS object is missing a required operation");
     return(KINLS_ILL_INPUT);
   }
 
+  /* Retrieve the LS type */
+  LSType = SUNLinSolGetType(LS);
+
+  /* Return with error if LS has 'matrix-embedded' type */
+  if (LSType == SUNLINEARSOLVER_MATRIX_EMBEDDED) {
+    KINProcessError(kin_mem, KINLS_ILL_INPUT, "KINLS",
+                   "KINSetLinearSolver",
+                   "KINSOL is incompatible with MATRIX_EMBEDDED LS objects");
+    return(KINLS_ILL_INPUT);
+  }
+
+  /* Set flags based on LS type */
+  iterative   = (LSType != SUNLINEARSOLVER_DIRECT);
+  matrixbased = (LSType != SUNLINEARSOLVER_ITERATIVE);
+
   /* check for required vector operations for KINLS interface */
   if ( (kin_mem->kin_vtemp1->ops->nvconst == NULL) ||
        (kin_mem->kin_vtemp1->ops->nvdotprod == NULL) ) {
@@ -81,32 +95,40 @@ int KINSetLinearSolver(void *kinmem, SUNLinearSolver LS, SUNMatrix A)
     return(KINLS_ILL_INPUT);
   }
 
-  /* Retrieve the LS type */
-  LSType = SUNLinSolGetType(LS);
-
   /* Check for compatible LS type, matrix and "atimes" support */
-  if ((LSType == SUNLINEARSOLVER_ITERATIVE) && (LS->ops->setatimes == NULL)) {
-    KINProcessError(kin_mem, KINLS_ILL_INPUT, "KINLS", "KINSetLinearSolver",
-                    "Incompatible inputs: iterative LS must support ATimes routine");
-    return(KINLS_ILL_INPUT);
-  }
-  if ((LSType == SUNLINEARSOLVER_DIRECT) && (A == NULL)) {
+  if (iterative) {
+
+    if ((LS->ops->setscalingvectors == NULL) &&
+        (kin_mem->kin_vtemp1->ops->nvgetlength == NULL)) {
+      KINProcessError(kin_mem, KINLS_ILL_INPUT, "KINLS",
+                      "KINSetLinearSolver", MSG_LS_BAD_NVECTOR);
+      return(KINLS_ILL_INPUT);
+    }
+
+    if (!matrixbased && (LS->ops->setatimes == NULL)) {
+      KINProcessError(kin_mem, KINLS_ILL_INPUT, "KINLS", "KINSetLinearSolver",
+                      "Incompatible inputs: iterative LS must support ATimes routine");
+      return(KINLS_ILL_INPUT);
+    }
+
+    if (matrixbased && A == NULL) {
+      KINProcessError(kin_mem, KINLS_ILL_INPUT, "KINLS", "KINSetLinearSolver",
+                      "Incompatible inputs: matrix-iterative LS requires non-NULL matrix");
+      return(KINLS_ILL_INPUT);
+    }
+
+  } else if (A == NULL) {
+
     KINProcessError(kin_mem, KINLS_ILL_INPUT, "KINLS", "KINSetLinearSolver",
                     "Incompatible inputs: direct LS requires non-NULL matrix");
     return(KINLS_ILL_INPUT);
   }
-  if ((LSType == SUNLINEARSOLVER_MATRIX_ITERATIVE) && (A == NULL)) {
-    KINProcessError(kin_mem, KINLS_ILL_INPUT, "KINLS", "KINSetLinearSolver",
-                    "Incompatible inputs: matrix-iterative LS requires non-NULL matrix");
-    return(KINLS_ILL_INPUT);
-  }
 
   /* free any existing system solver attached to KIN */
   if (kin_mem->kin_lfree) kin_mem->kin_lfree(kin_mem);
 
   /* Determine if this is an iterative linear solver */
-  kin_mem->kin_inexact_ls = ( (LSType == SUNLINEARSOLVER_ITERATIVE) ||
-                              (LSType == SUNLINEARSOLVER_MATRIX_ITERATIVE) );
+  kin_mem->kin_inexact_ls = iterative;
 
   /* Set four main system linear solver function fields in kin_mem */
   kin_mem->kin_linit  = kinLsInitialize;
@@ -139,6 +161,7 @@ int KINSetLinearSolver(void *kinmem, SUNLinearSolver LS, SUNMatrix A)
   }
   kinls_mem->jtimesDQ = SUNTRUE;
   kinls_mem->jtimes   = kinLsDQJtimes;
+  kinls_mem->jt_func  = kin_mem->kin_func;
   kinls_mem->jt_data  = kin_mem;
 
   /* Set defaults for preconditioner-related fields */
@@ -239,8 +262,8 @@ int KINSetPreconditioner(void *kinmem,
 {
   KINMem   kin_mem;
   KINLsMem kinls_mem;
-  PSetupFn kinls_psetup;
-  PSolveFn kinls_psolve;
+  SUNPSetupFn kinls_psetup;
+  SUNPSolveFn kinls_psolve;
   int      retval;
 
   /* access KINLsMem structure */
@@ -303,12 +326,44 @@ int KINSetJacTimesVecFn(void *kinmem, KINLsJacTimesVecFn jtv)
   } else {
     kinls_mem->jtimesDQ = SUNTRUE;
     kinls_mem->jtimes   = kinLsDQJtimes;
+    kinls_mem->jt_func  = kin_mem->kin_func;
     kinls_mem->jt_data  = kin_mem;
   }
 
   return(KINLS_SUCCESS);
 }
 
+
+/* KINSetJacTimesVecSysFn specifies an alternative user-supplied system function
+   to use in the internal finite difference Jacobian-vector product */
+int KINSetJacTimesVecSysFn(void *kinmem, KINSysFn jtimesSysFn)
+{
+  int      retval;
+  KINMem   kin_mem = NULL;
+  KINLsMem kinls_mem = NULL;
+
+  /* access KINLsMem structure */
+  retval = kinLs_AccessLMem(kin_mem, "KINSetJacTimesVecSysFn",
+                            &kin_mem, &kinls_mem);
+  if (retval != KIN_SUCCESS) return(retval);
+
+  /* check if using internal finite difference approximation */
+  if (!(kinls_mem->jtimesDQ)) {
+    KINProcessError(kin_mem, KINLS_ILL_INPUT, "KINLS", "KINSetJacTimesVecSysFn",
+                    "Internal finite-difference Jacobian-vector product is disabled.");
+    return(KINLS_ILL_INPUT);
+  }
+
+  /* store function pointers for system function (NULL implies use kin_func) */
+  if (jtimesSysFn != NULL)
+    kinls_mem->jt_func = jtimesSysFn;
+  else
+    kinls_mem->jt_func = kin_mem->kin_func;
+
+  return(KINLS_SUCCESS);
+}
+
+
 /*------------------------------------------------------------------
   KINGetLinWorkSpace returns the integer and real workspace size
   ------------------------------------------------------------------*/
@@ -713,7 +768,7 @@ int kinLsDenseDQJac(N_Vector u, N_Vector fu, SUNMatrix Jac,
   kinls_mem = (KINLsMem) kin_mem->kin_lmem;
 
   /* access matrix dimension */
-  N = SUNDenseMatrix_Rows(Jac);
+  N = SUNDenseMatrix_Columns(Jac);
 
   /* Save pointer to the array in tmp2 */
   tmp2_data = N_VGetArrayPointer(tmp2);
@@ -909,8 +964,8 @@ int kinLsDQJtimes(N_Vector v, N_Vector Jv, N_Vector u,
   N_VLinearSum(ONE, u, sigma, v, kin_mem->kin_vtemp1);
 
   /* call the system function to calculate func(u+sigma*v) */
-  retval = kin_mem->kin_func(kin_mem->kin_vtemp1, kin_mem->kin_vtemp2,
-                             kin_mem->kin_user_data);
+  retval = kinls_mem->jt_func(kin_mem->kin_vtemp1, kin_mem->kin_vtemp2,
+                              kin_mem->kin_user_data);
   kinls_mem->nfeDQ++;
   if (retval != 0) return(retval);
 
@@ -928,7 +983,7 @@ int kinLsDQJtimes(N_Vector v, N_Vector Jv, N_Vector u,
 int kinLsInitialize(KINMem kin_mem)
 {
   KINLsMem kinls_mem;
-  int      retval, LSType;
+  int      retval;
 
   /* Access KINLsMem structure */
   if (kin_mem->kin_lmem == NULL) {
@@ -938,9 +993,6 @@ int kinLsInitialize(KINMem kin_mem)
   }
   kinls_mem = (KINLsMem) kin_mem->kin_lmem;
 
-  /* Retrieve the LS type */
-  LSType = SUNLinSolGetType(kinls_mem->LS);
-
   /* Test for valid combinations of matrix & Jacobian routines: */
   if (kinls_mem->J == NULL) {
 
@@ -1047,17 +1099,11 @@ int kinLsInitialize(KINMem kin_mem)
        <=> \sum_{i=0}^{n-1} (b - A x_i)^2 < tol^2 / fs_mean^2
        <=> || b - A x ||_2 < tol / fs_mean
        <=> || b - A x ||_2 < tol * tol_fac
-     So we compute tol_fac = 1 / ||fscale||_RMS = sqrt(n) / ||fscale||_2,
-     for scaling desired tolerances */
-  if ( ((LSType == SUNLINEARSOLVER_ITERATIVE) ||
-        (LSType == SUNLINEARSOLVER_MATRIX_ITERATIVE)) &&
-       (kinls_mem->LS->ops->setscalingvectors == NULL) ) {
-
-    /* compute tol_fac = ||1||_2 / ||fscale||_2 */
+     So we compute tol_fac = sqrt(N) / ||fscale||_L2 for scaling desired tolerances */
+  if (kinls_mem->iterative && kinls_mem->LS->ops->setscalingvectors == NULL) {
     N_VConst(ONE, kin_mem->kin_vtemp1);
-    kinls_mem->tol_fac = SUNRsqrt( N_VDotProd(kin_mem->kin_vtemp1,
-                                              kin_mem->kin_vtemp1) )
-      / SUNRsqrt( N_VDotProd(kin_mem->kin_fscale, kin_mem->kin_fscale) );
+    kinls_mem->tol_fac = SUNRsqrt(N_VGetLength(kin_mem->kin_vtemp1))
+                       / N_VWL2Norm(kin_mem->kin_fscale, kin_mem->kin_vtemp1);
   } else {
     kinls_mem->tol_fac = ONE;
   }
@@ -1084,20 +1130,21 @@ int kinLsSetup(KINMem kin_mem)
   }
   kinls_mem = (KINLsMem) kin_mem->kin_lmem;
 
-
   /* recompute if J if it is non-NULL */
   if (kinls_mem->J) {
 
     /* Increment nje counter. */
     kinls_mem->nje++;
 
-    /* Zero out J */
-    retval = SUNMatZero(kinls_mem->J);
-    if (retval != 0) {
-      KINProcessError(kin_mem, KINLS_SUNMAT_FAIL, "KINLS",
-                      "kinLsSetup", MSG_LS_MATZERO_FAILED);
-      kinls_mem->last_flag = KINLS_SUNMAT_FAIL;
-      return(kinls_mem->last_flag);
+    /* Clear the linear system matrix if necessary */
+    if (SUNLinSolGetType(kinls_mem->LS) == SUNLINEARSOLVER_DIRECT) {
+      retval = SUNMatZero(kinls_mem->J);
+      if (retval != 0) {
+        KINProcessError(kin_mem, KINLS_SUNMAT_FAIL, "KINLS",
+                        "kinLsSetup", MSG_LS_MATZERO_FAILED);
+        kinls_mem->last_flag = KINLS_SUNMAT_FAIL;
+        return(kinls_mem->last_flag);
+      }
     }
 
     /* Call Jacobian routine */
@@ -1132,7 +1179,7 @@ int kinLsSolve(KINMem kin_mem, N_Vector xx, N_Vector bb,
 {
   KINLsMem kinls_mem;
   int      nli_inc, retval;
-  realtype res_norm, tol, LSType;
+  realtype res_norm, tol;
 
   /* Access KINLsMem structure */
   if (kin_mem->kin_lmem == NULL) {
@@ -1142,9 +1189,6 @@ int kinLsSolve(KINMem kin_mem, N_Vector xx, N_Vector bb,
   }
   kinls_mem = (KINLsMem) kin_mem->kin_lmem;
 
-  /* Retrieve the LS type */
-  LSType = SUNLinSolGetType(kinls_mem->LS);
-
   /* Set linear solver tolerance as input value times scaling factor
      (to account for possible lack of support for left/right scaling
      vectors in SUNLinSol object) */
@@ -1153,6 +1197,10 @@ int kinLsSolve(KINMem kin_mem, N_Vector xx, N_Vector bb,
   /* Set initial guess x = 0 to LS */
   N_VConst(ZERO, xx);
 
+  /* Set zero initial guess flag */
+  retval = SUNLinSolSetZeroGuess(kinls_mem->LS, SUNTRUE);
+  if (retval != SUNLS_SUCCESS) return(-1);
+
   /* set flag required for user-supplied J*v routine */
   kinls_mem->new_uu = SUNTRUE;
 
@@ -1167,9 +1215,7 @@ int kinLsSolve(KINMem kin_mem, N_Vector xx, N_Vector bb,
   if (kinls_mem->LS->ops->numiters)
     nli_inc = SUNLinSolNumIters(kinls_mem->LS);
 
-  if ( ((LSType == SUNLINEARSOLVER_ITERATIVE) ||
-        (LSType == SUNLINEARSOLVER_MATRIX_ITERATIVE)) &&
-       (kin_mem->kin_printfl > 2) )
+  if (kinls_mem->iterative && kin_mem->kin_printfl > 2)
     KINPrintInfo(kin_mem, PRNT_NLI, "KINLS", "kinLsSolve",
                  INFO_NLI, nli_inc);
 
@@ -1218,43 +1264,39 @@ int kinLsSolve(KINMem kin_mem, N_Vector xx, N_Vector bb,
     return(retval);
   }
 
-  /* SUNLinSolSolve returned SUNLS_SUCCESS or SUNLS_RES_REDUCED
+  /* SUNLinSolSolve returned SUNLS_SUCCESS or SUNLS_RES_REDUCED */
 
-     Compute auxiliary values for use in the linesearch and in KINForcingTerm.
+  /* Compute auxiliary values for use in the linesearch and in KINForcingTerm.
      These will be subsequently corrected if the step is reduced by constraints
      or the linesearch. */
+  if (kin_mem->kin_globalstrategy != KIN_FP) {
 
-
-  /* sJpnorm is the norm of the scaled product (scaled by fscale) of the
-     current Jacobian matrix J and the step vector p (= solution vector xx).
-
-     Only compute this if KINForcingTerm will eventually be called */
-  if ( (kin_mem->kin_globalstrategy != KIN_PICARD) &&
-       (kin_mem->kin_globalstrategy != KIN_FP) &&
-       (kin_mem->kin_callForcingTerm) ) {
-
-    retval = kinLsATimes(kin_mem, xx, bb);
-    if (retval > 0) {
-      kinls_mem->last_flag = SUNLS_ATIMES_FAIL_REC;
-      return(1);
+    /* sJpnorm is the norm of the scaled product (scaled by fscale) of the
+       current Jacobian matrix J and the step vector p (= solution vector xx) */
+    if (kin_mem->kin_inexact_ls && kin_mem->kin_etaflag == KIN_ETACHOICE1) {
+      retval = kinLsATimes(kin_mem, xx, bb);
+      if (retval > 0) {
+        kinls_mem->last_flag = SUNLS_ATIMES_FAIL_REC;
+        return(1);
+      }
+      else if (retval < 0) {
+        kinls_mem->last_flag = SUNLS_ATIMES_FAIL_UNREC;
+        return(-1);
+      }
+      *sJpnorm = N_VWL2Norm(bb, kin_mem->kin_fscale);
     }
-    else if (retval < 0) {
-      kinls_mem->last_flag = SUNLS_ATIMES_FAIL_UNREC;
-      return(-1);
-    }
-    *sJpnorm = N_VWL2Norm(bb, kin_mem->kin_fscale);
 
+    /* sFdotJp is the dot product of the scaled f vector and the scaled
+       vector J*p, where the scaling uses fscale */
+    if ((kin_mem->kin_inexact_ls && kin_mem->kin_etaflag == KIN_ETACHOICE1) ||
+        kin_mem->kin_globalstrategy == KIN_LINESEARCH) {
+      N_VProd(bb, kin_mem->kin_fscale, bb);
+      N_VProd(bb, kin_mem->kin_fscale, bb);
+      *sFdotJp = N_VDotProd(kin_mem->kin_fval, bb);
+    }
   }
 
-  /* sFdotJp is the dot product of the scaled f vector and the scaled
-     vector J*p, where the scaling uses fscale */
-  N_VProd(bb, kin_mem->kin_fscale, bb);
-  N_VProd(bb, kin_mem->kin_fscale, bb);
-  *sFdotJp = N_VDotProd(kin_mem->kin_fval, bb);
-
-  if ( ((LSType == SUNLINEARSOLVER_ITERATIVE) ||
-        (LSType == SUNLINEARSOLVER_MATRIX_ITERATIVE)) &&
-       (kin_mem->kin_printfl > 2) )
+  if (kin_mem->kin_inexact_ls && kin_mem->kin_printfl > 2)
     KINPrintInfo(kin_mem, PRNT_EPS, "KINLS", "kinLsSolve",
                  INFO_EPS, res_norm, kin_mem->kin_eps);
 
