@@ -1,8 +1,8 @@
 /* -----------------------------------------------------------------
- * Programmer(s): David J. Gardner @ LLNL                               
+ * Programmer(s): David J. Gardner @ LLNL
  * -----------------------------------------------------------------
  * SUNDIALS Copyright Start
- * Copyright (c) 2002-2019, Lawrence Livermore National Security
+ * Copyright (c) 2002-2022, Lawrence Livermore National Security
  * and Southern Methodist University.
  * All rights reserved.
  *
@@ -19,30 +19,36 @@
 
 #include <sundials/sundials_version.h>
 
+/* note strlen does not include terminating null character hence the
+   use of >= when checking len below and strncpy copies up to len
+   characters including the terminating null character */
+
 /* fill string with SUNDIALS version information */
 int SUNDIALSGetVersion(char *version, int len)
 {
-  if (strlen(SUNDIALS_VERSION) > len) {
-    return(-1);
-  }
-  
-  strncpy(version, SUNDIALS_VERSION, len);
+  if (version == NULL) return(-1);
+  if (strlen(SUNDIALS_VERSION) >= (size_t)len) return(-1);
+
+  strncpy(version, SUNDIALS_VERSION, (size_t)len);
+
   return(0);
 }
 
-/* fill integers with SUNDIALS major, minor, and patch release 
+/* fill integers with SUNDIALS major, minor, and patch release
    numbers and fill a string with the release label */
-int SUNDIALSGetVersionNumber(int *major, int *minor, int *patch, 
+int SUNDIALSGetVersionNumber(int *major, int *minor, int *patch,
                              char *label, int len)
 {
-  if (strlen(SUNDIALS_VERSION_LABEL) > len) {
-    return(-1);
-  }
-  
+  if (major == NULL ||
+      minor == NULL ||
+      patch == NULL ||
+      label == NULL) return(-1);
+  if (strlen(SUNDIALS_VERSION_LABEL) >= (size_t)len) return(-1);
+
   *major = SUNDIALS_VERSION_MAJOR;
   *minor = SUNDIALS_VERSION_MINOR;
   *patch = SUNDIALS_VERSION_PATCH;
-  strncpy(label, SUNDIALS_VERSION_LABEL, len);
+  strncpy(label, SUNDIALS_VERSION_LABEL, (size_t)len);
 
   return(0);
 }
