@@ -1,11 +1,10 @@
-/*
- * -----------------------------------------------------------------
+/* -----------------------------------------------------------------
  * Programmer(s): Daniel Reynolds @ SMU
- * Based on sundials_spbcgs.c code, written by Peter Brown and 
+ * Based on sundials_spbcgs.c code, written by Peter Brown and
  *                Aaron Collier @ LLNL
  * -----------------------------------------------------------------
  * SUNDIALS Copyright Start
- * Copyright (c) 2002-2019, Lawrence Livermore National Security
+ * Copyright (c) 2002-2022, Lawrence Livermore National Security
  * and Southern Methodist University.
  * All rights reserved.
  *
@@ -14,10 +13,9 @@
  * SPDX-License-Identifier: BSD-3-Clause
  * SUNDIALS Copyright End
  * -----------------------------------------------------------------
- * This is the implementation file for the SPBCGS implementation of 
+ * This is the implementation file for the SPBCGS implementation of
  * the SUNLINSOL package.
- * -----------------------------------------------------------------
- */ 
+ * -----------------------------------------------------------------*/
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -25,12 +23,14 @@
 #include <sunlinsol/sunlinsol_spbcgs.h>
 #include <sundials/sundials_math.h>
 
+#include "sundials_debug.h"
+
 #define ZERO RCONST(0.0)
 #define ONE  RCONST(1.0)
 
 /*
  * -----------------------------------------------------------------
- * SPBCGS solver structure accessibility macros: 
+ * SPBCGS solver structure accessibility macros:
  * -----------------------------------------------------------------
  */
 
@@ -38,21 +38,6 @@
 #define PRETYPE(S)         ( SPBCGS_CONTENT(S)->pretype )
 #define LASTFLAG(S)        ( SPBCGS_CONTENT(S)->last_flag )
 
-/*
- * -----------------------------------------------------------------
- * deprecated wrapper functions
- * -----------------------------------------------------------------
- */
-
-SUNLinearSolver SUNSPBCGS(N_Vector y, int pretype, int maxl)
-{ return(SUNLinSol_SPBCGS(y, pretype, maxl)); }
-
-int SUNSPBCGSSetPrecType(SUNLinearSolver S, int pretype)
-{ return(SUNLinSol_SPBCGSSetPrecType(S, pretype)); }
-
-int SUNSPBCGSSetMaxl(SUNLinearSolver S, int maxl)
-{ return(SUNLinSol_SPBCGSSetMaxl(S, maxl)); }
-
 /*
  * -----------------------------------------------------------------
  * exported functions
@@ -63,16 +48,15 @@ int SUNSPBCGSSetMaxl(SUNLinearSolver S, int maxl)
  * Function to create a new SPBCGS linear solver
  */
 
-SUNLinearSolver SUNLinSol_SPBCGS(N_Vector y, int pretype, int maxl)
+SUNLinearSolver SUNLinSol_SPBCGS(N_Vector y, int pretype, int maxl, SUNContext sunctx)
 {
   SUNLinearSolver S;
-  SUNLinearSolver_Ops ops;
   SUNLinearSolverContent_SPBCGS content;
-  
+
   /* check for legal pretype and maxl values; if illegal use defaults */
-  if ((pretype != PREC_NONE)  && (pretype != PREC_LEFT) &&
-      (pretype != PREC_RIGHT) && (pretype != PREC_BOTH))
-    pretype = PREC_NONE;
+  if ((pretype != SUN_PREC_NONE)  && (pretype != SUN_PREC_LEFT) &&
+      (pretype != SUN_PREC_RIGHT) && (pretype != SUN_PREC_BOTH))
+    pretype = SUN_PREC_NONE;
   if (maxl <= 0)
     maxl = SUNSPBCGS_MAXL_DEFAULT;
 
@@ -85,79 +69,93 @@ SUNLinearSolver SUNLinSol_SPBCGS(N_Vector y, int pretype, int maxl)
 
   /* Create linear solver */
   S = NULL;
-  S = (SUNLinearSolver) malloc(sizeof *S);
+  S = SUNLinSolNewEmpty(sunctx);
   if (S == NULL) return(NULL);
-  
-  /* Create linear solver operation structure */
-  ops = NULL;
-  ops = (SUNLinearSolver_Ops) malloc(sizeof(struct _generic_SUNLinearSolver_Ops));
-  if (ops == NULL) { free(S); return(NULL); }
 
   /* Attach operations */
-  ops->gettype           = SUNLinSolGetType_SPBCGS;
-  ops->setatimes         = SUNLinSolSetATimes_SPBCGS;
-  ops->setpreconditioner = SUNLinSolSetPreconditioner_SPBCGS;
-  ops->setscalingvectors = SUNLinSolSetScalingVectors_SPBCGS;
-  ops->initialize        = SUNLinSolInitialize_SPBCGS;
-  ops->setup             = SUNLinSolSetup_SPBCGS;
-  ops->solve             = SUNLinSolSolve_SPBCGS;
-  ops->numiters          = SUNLinSolNumIters_SPBCGS;
-  ops->resnorm           = SUNLinSolResNorm_SPBCGS;
-  ops->resid             = SUNLinSolResid_SPBCGS;
-  ops->lastflag          = SUNLinSolLastFlag_SPBCGS;  
-  ops->space             = SUNLinSolSpace_SPBCGS;  
-  ops->free              = SUNLinSolFree_SPBCGS;
+  S->ops->gettype           = SUNLinSolGetType_SPBCGS;
+  S->ops->getid             = SUNLinSolGetID_SPBCGS;
+  S->ops->setatimes         = SUNLinSolSetATimes_SPBCGS;
+  S->ops->setpreconditioner = SUNLinSolSetPreconditioner_SPBCGS;
+  S->ops->setscalingvectors = SUNLinSolSetScalingVectors_SPBCGS;
+  S->ops->setzeroguess      = SUNLinSolSetZeroGuess_SPBCGS;
+  S->ops->initialize        = SUNLinSolInitialize_SPBCGS;
+  S->ops->setup             = SUNLinSolSetup_SPBCGS;
+  S->ops->solve             = SUNLinSolSolve_SPBCGS;
+  S->ops->numiters          = SUNLinSolNumIters_SPBCGS;
+  S->ops->resnorm           = SUNLinSolResNorm_SPBCGS;
+  S->ops->resid             = SUNLinSolResid_SPBCGS;
+  S->ops->lastflag          = SUNLinSolLastFlag_SPBCGS;
+  S->ops->space             = SUNLinSolSpace_SPBCGS;
+  S->ops->free              = SUNLinSolFree_SPBCGS;
 
   /* Create content */
   content = NULL;
-  content = (SUNLinearSolverContent_SPBCGS) malloc(sizeof(struct _SUNLinearSolverContent_SPBCGS));
-  if (content == NULL) { free(ops); free(S); return(NULL); }
+  content = (SUNLinearSolverContent_SPBCGS) malloc(sizeof *content);
+  if (content == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  /* Attach content */
+  S->content = content;
 
   /* Fill content */
-  content->last_flag = 0;
-  content->maxl = maxl;
-  content->pretype = pretype;
-  content->numiters = 0;
-  content->resnorm = ZERO;
-  content->r_star = N_VClone(y);
-  if (content->r_star == NULL)  return(NULL);
-  content->r = N_VClone(y);
-  if (content->r == NULL)  return(NULL);
-  content->p = N_VClone(y);
-  if (content->p == NULL)  return(NULL);
-  content->q = N_VClone(y);
-  if (content->q == NULL)  return(NULL);
-  content->u = N_VClone(y);
-  if (content->u == NULL)  return(NULL);
-  content->Ap = N_VClone(y);
-  if (content->Ap == NULL)  return(NULL);
-  content->vtemp = N_VClone(y);
-  if (content->vtemp == NULL)  return(NULL);
-  content->s1 = NULL;
-  content->s2 = NULL;
-  content->ATimes = NULL;
-  content->ATData = NULL;
-  content->Psetup = NULL;
-  content->Psolve = NULL;
-  content->PData = NULL;
+  content->last_flag   = 0;
+  content->maxl        = maxl;
+  content->pretype     = pretype;
+  content->zeroguess   = SUNFALSE;
+  content->numiters    = 0;
+  content->resnorm     = ZERO;
+  content->r_star      = NULL;
+  content->r           = NULL;
+  content->p           = NULL;
+  content->q           = NULL;
+  content->u           = NULL;
+  content->Ap          = NULL;
+  content->vtemp       = NULL;
+  content->s1          = NULL;
+  content->s2          = NULL;
+  content->ATimes      = NULL;
+  content->ATData      = NULL;
+  content->Psetup      = NULL;
+  content->Psolve      = NULL;
+  content->PData       = NULL;
+  content->print_level = 0;
+  content->info_file   = stdout;
 
-  /* Attach content and ops */
-  S->content = content;
-  S->ops     = ops;
+  /* Allocate content */
+  content->r_star = N_VClone(y);
+  if (content->r_star == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->r = N_VClone(y);
+  if (content->r == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->p = N_VClone(y);
+  if (content->p == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->q = N_VClone(y);
+  if (content->q == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->u = N_VClone(y);
+  if (content->u == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->Ap = N_VClone(y);
+  if (content->Ap == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->vtemp = N_VClone(y);
+  if (content->vtemp == NULL) { SUNLinSolFree(S); return(NULL); }
 
   return(S);
 }
 
 
 /* ----------------------------------------------------------------------------
- * Function to set the type of preconditioning for SPBCGS to use 
+ * Function to set the type of preconditioning for SPBCGS to use
  */
 
-SUNDIALS_EXPORT int SUNLinSol_SPBCGSSetPrecType(SUNLinearSolver S, int pretype) 
+int SUNLinSol_SPBCGSSetPrecType(SUNLinearSolver S, int pretype)
 {
-  /* Check for legal pretype */ 
-  if ((pretype != PREC_NONE)  && (pretype != PREC_LEFT) &&
-      (pretype != PREC_RIGHT) && (pretype != PREC_BOTH)) {
+  /* Check for legal pretype */
+  if ((pretype != SUN_PREC_NONE)  && (pretype != SUN_PREC_LEFT) &&
+      (pretype != SUN_PREC_RIGHT) && (pretype != SUN_PREC_BOTH)) {
     return(SUNLS_ILL_INPUT);
   }
 
@@ -171,15 +169,15 @@ SUNDIALS_EXPORT int SUNLinSol_SPBCGSSetPrecType(SUNLinearSolver S, int pretype)
 
 
 /* ----------------------------------------------------------------------------
- * Function to set the maximum number of iterations for SPBCGS to use 
+ * Function to set the maximum number of iterations for SPBCGS to use
  */
 
-SUNDIALS_EXPORT int SUNLinSol_SPBCGSSetMaxl(SUNLinearSolver S, int maxl) 
+int SUNLinSol_SPBCGSSetMaxl(SUNLinearSolver S, int maxl)
 {
   /* Check for non-NULL SUNLinearSolver */
   if (S == NULL) return(SUNLS_MEM_NULL);
 
-  /* Check for legal pretype */ 
+  /* Check for legal pretype */
   if (maxl <= 0)
     maxl = SUNSPBCGS_MAXL_DEFAULT;
 
@@ -201,27 +199,45 @@ SUNLinearSolver_Type SUNLinSolGetType_SPBCGS(SUNLinearSolver S)
 }
 
 
+SUNLinearSolver_ID SUNLinSolGetID_SPBCGS(SUNLinearSolver S)
+{
+  return(SUNLINEARSOLVER_SPBCGS);
+}
+
+
 int SUNLinSolInitialize_SPBCGS(SUNLinearSolver S)
 {
   /* ensure valid options */
-  if (S == NULL) return(SUNLS_MEM_NULL);  
-  if ( (PRETYPE(S) != PREC_LEFT) && 
-       (PRETYPE(S) != PREC_RIGHT) && 
-       (PRETYPE(S) != PREC_BOTH) )
-    PRETYPE(S) = PREC_NONE;
-  if (SPBCGS_CONTENT(S)->maxl <= 0) 
+  if (S == NULL) return(SUNLS_MEM_NULL);
+
+  if (SPBCGS_CONTENT(S)->maxl <= 0)
     SPBCGS_CONTENT(S)->maxl = SUNSPBCGS_MAXL_DEFAULT;
 
+  if (SPBCGS_CONTENT(S)->ATimes == NULL) {
+    LASTFLAG(S) = SUNLS_ATIMES_NULL;
+    return(LASTFLAG(S));
+  }
+
+  if ( (PRETYPE(S) != SUN_PREC_LEFT) &&
+       (PRETYPE(S) != SUN_PREC_RIGHT) &&
+       (PRETYPE(S) != SUN_PREC_BOTH) )
+    PRETYPE(S) = SUN_PREC_NONE;
+
+  if ((PRETYPE(S) != SUN_PREC_NONE) && (SPBCGS_CONTENT(S)->Psolve == NULL)) {
+    LASTFLAG(S) = SUNLS_PSOLVE_NULL;
+    return(LASTFLAG(S));
+  }
+
   /* no additional memory to allocate */
-  
+
   /* return with success */
   LASTFLAG(S) = SUNLS_SUCCESS;
   return(LASTFLAG(S));
 }
 
 
-int SUNLinSolSetATimes_SPBCGS(SUNLinearSolver S, void* ATData, 
-                              ATimesFn ATimes)
+int SUNLinSolSetATimes_SPBCGS(SUNLinearSolver S, void* ATData,
+                              SUNATimesFn ATimes)
 {
   /* set function pointers to integrator-supplied ATimes routine
      and data, and return with success */
@@ -234,7 +250,7 @@ int SUNLinSolSetATimes_SPBCGS(SUNLinearSolver S, void* ATData,
 
 
 int SUNLinSolSetPreconditioner_SPBCGS(SUNLinearSolver S, void* PData,
-                                      PSetupFn Psetup, PSolveFn Psolve)
+                                      SUNPSetupFn Psetup, SUNPSolveFn Psolve)
 {
   /* set function pointers to integrator-supplied Psetup and PSolve
      routines and data, and return with success */
@@ -250,7 +266,7 @@ int SUNLinSolSetPreconditioner_SPBCGS(SUNLinearSolver S, void* PData,
 int SUNLinSolSetScalingVectors_SPBCGS(SUNLinearSolver S, N_Vector s1,
                                       N_Vector s2)
 {
-  /* set N_Vector pointers to integrator-supplied scaling vectors, 
+  /* set N_Vector pointers to integrator-supplied scaling vectors,
      and return with success */
   if (S == NULL) return(SUNLS_MEM_NULL);
   SPBCGS_CONTENT(S)->s1 = s1;
@@ -260,53 +276,64 @@ int SUNLinSolSetScalingVectors_SPBCGS(SUNLinearSolver S, N_Vector s1,
 }
 
 
+int SUNLinSolSetZeroGuess_SPBCGS(SUNLinearSolver S, booleantype onoff)
+{
+  /* set flag indicating a zero initial guess */
+  if (S == NULL) return(SUNLS_MEM_NULL);
+  SPBCGS_CONTENT(S)->zeroguess = onoff;
+  LASTFLAG(S) = SUNLS_SUCCESS;
+  return(LASTFLAG(S));
+}
+
+
 int SUNLinSolSetup_SPBCGS(SUNLinearSolver S, SUNMatrix A)
 {
   int ier;
-  PSetupFn Psetup;
+  SUNPSetupFn Psetup;
   void* PData;
 
   /* Set shortcuts to SPBCGS memory structures */
   if (S == NULL) return(SUNLS_MEM_NULL);
   Psetup = SPBCGS_CONTENT(S)->Psetup;
   PData = SPBCGS_CONTENT(S)->PData;
-  
-  /* no solver-specific setup is required, but if user-supplied 
+
+  /* no solver-specific setup is required, but if user-supplied
      Psetup routine exists, call that here */
   if (Psetup != NULL) {
     ier = Psetup(PData);
     if (ier != 0) {
-      LASTFLAG(S) = (ier < 0) ? 
-	SUNLS_PSET_FAIL_UNREC : SUNLS_PSET_FAIL_REC;
+      LASTFLAG(S) = (ier < 0) ?
+        SUNLS_PSET_FAIL_UNREC : SUNLS_PSET_FAIL_REC;
       return(LASTFLAG(S));
     }
   }
-  
-  /* return with success */ 
+
+  /* return with success */
   LASTFLAG(S) = SUNLS_SUCCESS;
   return(LASTFLAG(S));
 }
 
 
-int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x, 
+int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
                           N_Vector b, realtype delta)
 {
   /* local data and shortcut variables */
   realtype alpha, beta, omega, omega_denom, beta_num, beta_denom, r_norm, rho;
   N_Vector r_star, r, p, q, u, Ap, vtemp;
   booleantype preOnLeft, preOnRight, scale_x, scale_b, converged;
+  booleantype *zeroguess;
   int l, l_max, ier;
   void *A_data, *P_data;
   N_Vector sx, sb;
-  ATimesFn atimes;
-  PSolveFn psolve;
+  SUNATimesFn atimes;
+  SUNPSolveFn psolve;
   realtype *res_norm;
   int *nli;
 
   /* local variables for fused vector operations */
   realtype cv[3];
   N_Vector Xv[3];
-  
+
   /* Make local shorcuts to solver variables. */
   if (S == NULL) return(SUNLS_MEM_NULL);
   l_max        = SPBCGS_CONTENT(S)->maxl;
@@ -323,6 +350,7 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   P_data       = SPBCGS_CONTENT(S)->PData;
   atimes       = SPBCGS_CONTENT(S)->ATimes;
   psolve       = SPBCGS_CONTENT(S)->Psolve;
+  zeroguess    = &(SPBCGS_CONTENT(S)->zeroguess);
   nli          = &(SPBCGS_CONTENT(S)->numiters);
   res_norm     = &(SPBCGS_CONTENT(S)->resnorm);
 
@@ -331,19 +359,47 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   converged = SUNFALSE;
 
   /* set booleantype flags for internal solver options */
-  preOnLeft  = ( (PRETYPE(S) == PREC_LEFT) || 
-                 (PRETYPE(S) == PREC_BOTH) );
-  preOnRight = ( (PRETYPE(S) == PREC_RIGHT) || 
-                 (PRETYPE(S) == PREC_BOTH) );
+  preOnLeft  = ( (PRETYPE(S) == SUN_PREC_LEFT) ||
+                 (PRETYPE(S) == SUN_PREC_BOTH) );
+  preOnRight = ( (PRETYPE(S) == SUN_PREC_RIGHT) ||
+                 (PRETYPE(S) == SUN_PREC_BOTH) );
   scale_x = (sx != NULL);
   scale_b = (sb != NULL);
 
+  /* Check for unsupported use case */
+  if (preOnRight && !(*zeroguess)) {
+    *zeroguess  = SUNFALSE;
+    LASTFLAG(S) = SUNLS_ILL_INPUT;
+    return(SUNLS_ILL_INPUT);
+  }
+
+#ifdef SUNDIALS_BUILD_WITH_MONITORING
+  if (SPBCGS_CONTENT(S)->print_level && SPBCGS_CONTENT(S)->info_file)
+    STAN_SUNDIALS_FPRINTF(SPBCGS_CONTENT(S)->info_file, "SUNLINSOL_SPBCGS:\n");
+#endif
+
+  /* Check if Atimes function has been set */
+  if (atimes == NULL) {
+    *zeroguess  = SUNFALSE;
+    LASTFLAG(S) = SUNLS_ATIMES_NULL;
+    return(LASTFLAG(S));
+  }
+
+  /* If preconditioning, check if psolve has been set */
+  if ((preOnLeft || preOnRight) && psolve == NULL) {
+    *zeroguess  = SUNFALSE;
+    LASTFLAG(S) = SUNLS_PSOLVE_NULL;
+    return(LASTFLAG(S));
+  }
+
   /* Set r_star to initial (unscaled) residual r_0 = b - A*x_0 */
 
-  if (N_VDotProd(x, x) == ZERO) N_VScale(ONE, b, r_star);
-  else {
+  if (*zeroguess) {
+    N_VScale(ONE, b, r_star);
+  } else {
     ier = atimes(A_data, x, r_star);
     if (ier != 0) {
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = (ier < 0) ?
         SUNLS_ATIMES_FAIL_UNREC : SUNLS_ATIMES_FAIL_REC;
       return(LASTFLAG(S));
@@ -354,8 +410,9 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   /* Apply left preconditioner and b-scaling to r_star = r_0 */
 
   if (preOnLeft) {
-    ier = psolve(P_data, r_star, r, delta, PREC_LEFT);
+    ier = psolve(P_data, r_star, r, delta, SUN_PREC_LEFT);
     if (ier != 0) {
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = (ier < 0) ?
         SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
       return(LASTFLAG(S));
@@ -374,7 +431,19 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
      return if small */
 
   *res_norm = r_norm = rho = SUNRsqrt(beta_denom);
+
+#ifdef SUNDIALS_BUILD_WITH_MONITORING
+  /* print the initial residual */
+  if (SPBCGS_CONTENT(S)->print_level && SPBCGS_CONTENT(S)->info_file)
+  {
+    STAN_SUNDIALS_FPRINTF(SPBCGS_CONTENT(S)->info_file,
+            SUNLS_MSG_RESIDUAL,
+            (long int) 0, *res_norm);
+  }
+#endif
+
   if (r_norm <= delta) {
+    *zeroguess  = SUNFALSE;
     LASTFLAG(S) = SUNLS_SUCCESS;
     return(LASTFLAG(S));
   }
@@ -384,6 +453,9 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   N_VScale(ONE, r_star, r);
   N_VScale(ONE, r_star, p);
 
+  /* Set x = sx x if non-zero guess */
+  if (scale_x && !(*zeroguess)) N_VProd(sx, x, x);
+
   /* Begin main iteration loop */
 
   for(l = 0; l < l_max; l++) {
@@ -401,8 +473,9 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
 
     if (preOnRight) {
       N_VScale(ONE, vtemp, Ap);
-      ier = psolve(P_data, Ap, vtemp, delta, PREC_RIGHT);
+      ier = psolve(P_data, Ap, vtemp, delta, SUN_PREC_RIGHT);
       if (ier != 0) {
+        *zeroguess  = SUNFALSE;
         LASTFLAG(S) = (ier < 0) ?
           SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
         return(LASTFLAG(S));
@@ -413,6 +486,7 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
 
     ier = atimes(A_data, vtemp, Ap );
     if (ier != 0) {
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = (ier < 0) ?
         SUNLS_ATIMES_FAIL_UNREC : SUNLS_ATIMES_FAIL_REC;
       return(LASTFLAG(S));
@@ -421,8 +495,9 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
     /*   Apply left preconditioner: vtemp = P1_inv A P2_inv sx_inv p */
 
     if (preOnLeft) {
-      ier = psolve(P_data, Ap, vtemp, delta, PREC_LEFT);
+      ier = psolve(P_data, Ap, vtemp, delta, SUN_PREC_LEFT);
       if (ier != 0) {
+        *zeroguess  = SUNFALSE;
         LASTFLAG(S) = (ier < 0) ?
           SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
         return(LASTFLAG(S));
@@ -455,8 +530,9 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
 
     if (preOnRight) {
       N_VScale(ONE, vtemp, u);
-      ier = psolve(P_data, u, vtemp, delta, PREC_RIGHT);
+      ier = psolve(P_data, u, vtemp, delta, SUN_PREC_RIGHT);
       if (ier != 0) {
+        *zeroguess  = SUNFALSE;
         LASTFLAG(S) = (ier < 0) ?
           SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
         return(LASTFLAG(S));
@@ -467,6 +543,7 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
 
     ier = atimes(A_data, vtemp, u );
     if (ier != 0) {
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = (ier < 0) ?
         SUNLS_ATIMES_FAIL_UNREC : SUNLS_ATIMES_FAIL_REC;
       return(LASTFLAG(S));
@@ -475,8 +552,9 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
     /*   Apply left preconditioner: vtemp = P1_inv A P2_inv sx_inv p */
 
     if (preOnLeft) {
-      ier = psolve(P_data, u, vtemp, delta, PREC_LEFT);
+      ier = psolve(P_data, u, vtemp, delta, SUN_PREC_LEFT);
       if (ier != 0) {
+        *zeroguess  = SUNFALSE;
         LASTFLAG(S) = (ier < 0) ?
           SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
         return(LASTFLAG(S));
@@ -497,17 +575,26 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
     omega = (N_VDotProd(u, q) / omega_denom);
 
     /* Update x = x + alpha*p + omega*q */
-    cv[0] = ONE;
-    Xv[0] = x;
+    if (l == 0 && *zeroguess) {
+      N_VLinearSum(alpha, p, omega, q, x);
+    } else {
+      cv[0] = ONE;
+      Xv[0] = x;
 
-    cv[1] = alpha;
-    Xv[1] = p;
+      cv[1] = alpha;
+      Xv[1] = p;
 
-    cv[2] = omega;
-    Xv[2] = q;
+      cv[2] = omega;
+      Xv[2] = q;
 
-    ier = N_VLinearCombination(3, cv, Xv, x);
-    if (ier != SUNLS_SUCCESS) return(SUNLS_VECTOROP_ERR);
+      ier = N_VLinearCombination(3, cv, Xv, x);
+      if (ier != SUNLS_SUCCESS) {
+        *zeroguess  = SUNFALSE;
+        LASTFLAG(S) = SUNLS_VECTOROP_ERR;
+        return(SUNLS_VECTOROP_ERR);
+      }
+
+    }
 
     /* Update the residual r = q - omega*u */
 
@@ -516,6 +603,17 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
     /* Set rho = norm(r) and check convergence */
 
     *res_norm = rho = SUNRsqrt(N_VDotProd(r, r));
+
+#ifdef SUNDIALS_BUILD_WITH_MONITORING
+    /* print current iteration number and the residual */
+    if (SPBCGS_CONTENT(S)->print_level && SPBCGS_CONTENT(S)->info_file)
+    {
+      STAN_SUNDIALS_FPRINTF(SPBCGS_CONTENT(S)->info_file,
+              SUNLS_MSG_RESIDUAL,
+              (long int) *nli, *res_norm);
+    }
+#endif
+
     if (rho <= delta) {
       converged = SUNTRUE;
       break;
@@ -538,7 +636,11 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
     Xv[2] = r;
 
     ier = N_VLinearCombination(3, cv, Xv, p);
-    if (ier != SUNLS_SUCCESS) return(SUNLS_VECTOROP_ERR);
+    if (ier != SUNLS_SUCCESS) {
+      *zeroguess  = SUNFALSE;
+      LASTFLAG(S) = SUNLS_VECTOROP_ERR;
+      return(SUNLS_VECTOROP_ERR);
+    }
 
     /* udpate beta_denom for next iteration */
     beta_denom = beta_num;
@@ -552,8 +654,9 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
 
     if (scale_x) N_VDiv(x, sx, x);
     if (preOnRight) {
-      ier = psolve(P_data, x, vtemp, delta, PREC_RIGHT);
+      ier = psolve(P_data, x, vtemp, delta, SUN_PREC_RIGHT);
       if (ier != 0) {
+        *zeroguess  = SUNFALSE;
         LASTFLAG(S) = (ier < 0) ?
           SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
         return(LASTFLAG(S));
@@ -561,14 +664,16 @@ int SUNLinSolSolve_SPBCGS(SUNLinearSolver S, SUNMatrix A, N_Vector x,
       N_VScale(ONE, vtemp, x);
     }
 
-    if (converged == SUNTRUE) 
+    *zeroguess = SUNFALSE;
+    if (converged == SUNTRUE)
       LASTFLAG(S) = SUNLS_SUCCESS;
-    else 
+    else
       LASTFLAG(S) = SUNLS_RES_REDUCED;
     return(LASTFLAG(S));
-    
+
   }
   else {
+    *zeroguess  = SUNFALSE;
     LASTFLAG(S) = SUNLS_CONV_FAIL;
     return(LASTFLAG(S));
   }
@@ -598,7 +703,7 @@ N_Vector SUNLinSolResid_SPBCGS(SUNLinearSolver S)
 }
 
 
-long int SUNLinSolLastFlag_SPBCGS(SUNLinearSolver S)
+sunindextype SUNLinSolLastFlag_SPBCGS(SUNLinearSolver S)
 {
   /* return the stored 'last_flag' value */
   if (S == NULL) return(-1);
@@ -606,8 +711,8 @@ long int SUNLinSolLastFlag_SPBCGS(SUNLinearSolver S)
 }
 
 
-int SUNLinSolSpace_SPBCGS(SUNLinearSolver S, 
-                          long int *lenrwLS, 
+int SUNLinSolSpace_SPBCGS(SUNLinearSolver S,
+                          long int *lenrwLS,
                           long int *leniwLS)
 {
   sunindextype liw1, lrw1;
@@ -625,25 +730,79 @@ int SUNLinSolFree_SPBCGS(SUNLinearSolver S)
 {
   if (S == NULL) return(SUNLS_SUCCESS);
 
-  /* delete items from within the content structure */
-  if (SPBCGS_CONTENT(S)->r_star)
-    N_VDestroy(SPBCGS_CONTENT(S)->r_star);
-  if (SPBCGS_CONTENT(S)->r)
-    N_VDestroy(SPBCGS_CONTENT(S)->r);
-  if (SPBCGS_CONTENT(S)->p)
-    N_VDestroy(SPBCGS_CONTENT(S)->p);
-  if (SPBCGS_CONTENT(S)->q)
-    N_VDestroy(SPBCGS_CONTENT(S)->q);
-  if (SPBCGS_CONTENT(S)->u)
-    N_VDestroy(SPBCGS_CONTENT(S)->u);
-  if (SPBCGS_CONTENT(S)->Ap)
-    N_VDestroy(SPBCGS_CONTENT(S)->Ap);
-  if (SPBCGS_CONTENT(S)->vtemp)
-    N_VDestroy(SPBCGS_CONTENT(S)->vtemp);
-
-  /* delete generic structures */
-  free(S->content);  S->content = NULL;
-  free(S->ops);  S->ops = NULL;
+  if (S->content) {
+    /* delete items from within the content structure */
+    if (SPBCGS_CONTENT(S)->r_star) {
+      N_VDestroy(SPBCGS_CONTENT(S)->r_star);
+      SPBCGS_CONTENT(S)->r_star = NULL;
+    }
+    if (SPBCGS_CONTENT(S)->r) {
+      N_VDestroy(SPBCGS_CONTENT(S)->r);
+      SPBCGS_CONTENT(S)->r = NULL;
+    }
+    if (SPBCGS_CONTENT(S)->p) {
+      N_VDestroy(SPBCGS_CONTENT(S)->p);
+      SPBCGS_CONTENT(S)->p = NULL;
+    }
+    if (SPBCGS_CONTENT(S)->q) {
+      N_VDestroy(SPBCGS_CONTENT(S)->q);
+      SPBCGS_CONTENT(S)->q = NULL;
+    }
+    if (SPBCGS_CONTENT(S)->u) {
+      N_VDestroy(SPBCGS_CONTENT(S)->u);
+      SPBCGS_CONTENT(S)->u = NULL;
+    }
+    if (SPBCGS_CONTENT(S)->Ap) {
+      N_VDestroy(SPBCGS_CONTENT(S)->Ap);
+      SPBCGS_CONTENT(S)->Ap = NULL;
+    }
+    if (SPBCGS_CONTENT(S)->vtemp) {
+      N_VDestroy(SPBCGS_CONTENT(S)->vtemp);
+      SPBCGS_CONTENT(S)->vtemp = NULL;
+    }
+    free(S->content); S->content = NULL;
+  }
+  if (S->ops) { free(S->ops); S->ops = NULL; }
   free(S); S = NULL;
   return(SUNLS_SUCCESS);
 }
+
+
+int SUNLinSolSetInfoFile_SPBCGS(SUNLinearSolver S,
+                                FILE* info_file)
+{
+#ifdef SUNDIALS_BUILD_WITH_MONITORING
+  /* check that the linear solver is non-null */
+  if (S == NULL)
+    return(SUNLS_MEM_NULL);
+
+  SPBCGS_CONTENT(S)->info_file = info_file;
+
+  return(SUNLS_SUCCESS);
+#else
+  SUNDIALS_DEBUG_PRINT("ERROR in SUNLinSolSetInfoFile_SPBCGS: SUNDIALS was not built with monitoring\n");
+  return(SUNLS_ILL_INPUT);
+#endif
+}
+
+
+int SUNLinSolSetPrintLevel_SPBCGS(SUNLinearSolver S,
+                                  int print_level)
+{
+#ifdef SUNDIALS_BUILD_WITH_MONITORING
+  /* check that the linear solver is non-null */
+  if (S == NULL)
+    return(SUNLS_MEM_NULL);
+
+  /* check for valid print level */
+  if (print_level < 0 || print_level > 1)
+    return(SUNLS_ILL_INPUT);
+
+  SPBCGS_CONTENT(S)->print_level = print_level;
+
+  return(SUNLS_SUCCESS);
+#else
+  SUNDIALS_DEBUG_PRINT("ERROR in SUNLinSolSetPrintLevel_SPBCGS: SUNDIALS was not built with monitoring\n");
+  return(SUNLS_ILL_INPUT);
+#endif
+}
