@@ -1,10 +1,9 @@
-/*
- * -----------------------------------------------------------------
+/* -----------------------------------------------------------------
  * Programmer(s): Daniel Reynolds @ SMU
  * Based on sundials_sptfqmr.c code, written by Aaron Collier @ LLNL
  * -----------------------------------------------------------------
  * SUNDIALS Copyright Start
- * Copyright (c) 2002-2019, Lawrence Livermore National Security
+ * Copyright (c) 2002-2022, Lawrence Livermore National Security
  * and Southern Methodist University.
  * All rights reserved.
  *
@@ -13,10 +12,9 @@
  * SPDX-License-Identifier: BSD-3-Clause
  * SUNDIALS Copyright End
  * -----------------------------------------------------------------
- * This is the implementation file for the SPTFQMR implementation of 
+ * This is the implementation file for the SPTFQMR implementation of
  * the SUNLINSOL package.
- * -----------------------------------------------------------------
- */ 
+ * -----------------------------------------------------------------*/
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -24,35 +22,20 @@
 #include <sunlinsol/sunlinsol_sptfqmr.h>
 #include <sundials/sundials_math.h>
 
+#include "sundials_debug.h"
+
 #define ZERO RCONST(0.0)
 #define ONE  RCONST(1.0)
 
 /*
  * -----------------------------------------------------------------
- * SPTFQMR solver structure accessibility macros: 
+ * SPTFQMR solver structure accessibility macros:
  * -----------------------------------------------------------------
  */
 
-  
 #define SPTFQMR_CONTENT(S)  ( (SUNLinearSolverContent_SPTFQMR)(S->content) )
 #define LASTFLAG(S)         ( SPTFQMR_CONTENT(S)->last_flag )
 
-
-/*
- * -----------------------------------------------------------------
- * deprecated wrapper functions
- * -----------------------------------------------------------------
- */
-
-SUNLinearSolver SUNSPTFQMR(N_Vector y, int pretype, int maxl)
-{ return(SUNLinSol_SPTFQMR(y, pretype, maxl)); }
-
-int SUNSPTFQMRSetPrecType(SUNLinearSolver S, int pretype)
-{ return(SUNLinSol_SPTFQMRSetPrecType(S, pretype)); }
-
-int SUNSPTFQMRSetMaxl(SUNLinearSolver S, int maxl)
-{ return(SUNLinSol_SPTFQMRSetMaxl(S, maxl)); }
-
 /*
  * -----------------------------------------------------------------
  * exported functions
@@ -63,16 +46,15 @@ int SUNSPTFQMRSetMaxl(SUNLinearSolver S, int maxl)
  * Function to create a new SPTFQMR linear solver
  */
 
-SUNLinearSolver SUNLinSol_SPTFQMR(N_Vector y, int pretype, int maxl)
+SUNLinearSolver SUNLinSol_SPTFQMR(N_Vector y, int pretype, int maxl, SUNContext sunctx)
 {
   SUNLinearSolver S;
-  SUNLinearSolver_Ops ops;
   SUNLinearSolverContent_SPTFQMR content;
-  
+
   /* check for legal pretype and maxl values; if illegal use defaults */
-  if ((pretype != PREC_NONE)  && (pretype != PREC_LEFT) &&
-      (pretype != PREC_RIGHT) && (pretype != PREC_BOTH))
-    pretype = PREC_NONE;
+  if ((pretype != SUN_PREC_NONE)  && (pretype != SUN_PREC_LEFT) &&
+      (pretype != SUN_PREC_RIGHT) && (pretype != SUN_PREC_BOTH))
+    pretype = SUN_PREC_NONE;
   if (maxl <= 0)
     maxl = SUNSPTFQMR_MAXL_DEFAULT;
 
@@ -85,85 +67,105 @@ SUNLinearSolver SUNLinSol_SPTFQMR(N_Vector y, int pretype, int maxl)
 
   /* Create linear solver */
   S = NULL;
-  S = (SUNLinearSolver) malloc(sizeof *S);
+  S = SUNLinSolNewEmpty(sunctx);
   if (S == NULL) return(NULL);
-  
-  /* Create linear solver operation structure */
-  ops = NULL;
-  ops = (SUNLinearSolver_Ops) malloc(sizeof(struct _generic_SUNLinearSolver_Ops));
-  if (ops == NULL) { free(S); return(NULL); }
 
   /* Attach operations */
-  ops->gettype           = SUNLinSolGetType_SPTFQMR;
-  ops->setatimes         = SUNLinSolSetATimes_SPTFQMR;
-  ops->setpreconditioner = SUNLinSolSetPreconditioner_SPTFQMR;
-  ops->setscalingvectors = SUNLinSolSetScalingVectors_SPTFQMR;
-  ops->initialize        = SUNLinSolInitialize_SPTFQMR;
-  ops->setup             = SUNLinSolSetup_SPTFQMR;
-  ops->solve             = SUNLinSolSolve_SPTFQMR;
-  ops->numiters          = SUNLinSolNumIters_SPTFQMR;
-  ops->resnorm           = SUNLinSolResNorm_SPTFQMR;
-  ops->resid             = SUNLinSolResid_SPTFQMR;
-  ops->lastflag          = SUNLinSolLastFlag_SPTFQMR;  
-  ops->space             = SUNLinSolSpace_SPTFQMR;  
-  ops->free              = SUNLinSolFree_SPTFQMR;
+  S->ops->gettype           = SUNLinSolGetType_SPTFQMR;
+  S->ops->getid             = SUNLinSolGetID_SPTFQMR;
+  S->ops->setatimes         = SUNLinSolSetATimes_SPTFQMR;
+  S->ops->setpreconditioner = SUNLinSolSetPreconditioner_SPTFQMR;
+  S->ops->setscalingvectors = SUNLinSolSetScalingVectors_SPTFQMR;
+  S->ops->setzeroguess      = SUNLinSolSetZeroGuess_SPTFQMR;
+  S->ops->initialize        = SUNLinSolInitialize_SPTFQMR;
+  S->ops->setup             = SUNLinSolSetup_SPTFQMR;
+  S->ops->solve             = SUNLinSolSolve_SPTFQMR;
+  S->ops->numiters          = SUNLinSolNumIters_SPTFQMR;
+  S->ops->resnorm           = SUNLinSolResNorm_SPTFQMR;
+  S->ops->resid             = SUNLinSolResid_SPTFQMR;
+  S->ops->lastflag          = SUNLinSolLastFlag_SPTFQMR;
+  S->ops->space             = SUNLinSolSpace_SPTFQMR;
+  S->ops->free              = SUNLinSolFree_SPTFQMR;
 
   /* Create content */
   content = NULL;
-  content = (SUNLinearSolverContent_SPTFQMR) malloc(sizeof(struct _SUNLinearSolverContent_SPTFQMR));
-  if (content == NULL) { free(ops); free(S); return(NULL); }
+  content = (SUNLinearSolverContent_SPTFQMR) malloc(sizeof *content);
+  if (content == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  /* Attach content */
+  S->content = content;
 
   /* Fill content */
-  content->last_flag = 0;
-  content->maxl = maxl;
-  content->pretype = pretype;
-  content->numiters = 0;
-  content->resnorm = ZERO;
-  content->r_star = N_VClone(y);
-  if (content->r_star == NULL)  return(NULL);
-  content->q = N_VClone(y);
-  if (content->q == NULL)  return(NULL);
-  content->d = N_VClone(y);
-  if (content->d == NULL)  return(NULL);
-  content->v = N_VClone(y);
-  if (content->v == NULL)  return(NULL);
-  content->p = N_VClone(y);
-  if (content->p == NULL)  return(NULL);
-  content->r = N_VCloneVectorArray(2, y);
-  if (content->r == NULL)  return(NULL);
-  content->u = N_VClone(y);
-  if (content->u == NULL)  return(NULL);
-  content->vtemp1 = N_VClone(y);
-  if (content->vtemp1 == NULL)  return(NULL);
-  content->vtemp2 = N_VClone(y);
-  if (content->vtemp2 == NULL)  return(NULL);
-  content->vtemp3 = N_VClone(y);
-  if (content->vtemp3 == NULL)  return(NULL);
-  content->s1 = NULL;
-  content->s2 = NULL;
-  content->ATimes = NULL;
-  content->ATData = NULL;
-  content->Psetup = NULL;
-  content->Psolve = NULL;
-  content->PData = NULL;
+  content->last_flag   = 0;
+  content->maxl        = maxl;
+  content->pretype     = pretype;
+  content->zeroguess   = SUNFALSE;
+  content->numiters    = 0;
+  content->resnorm     = ZERO;
+  content->r_star      = NULL;
+  content->q           = NULL;
+  content->d           = NULL;
+  content->v           = NULL;
+  content->p           = NULL;
+  content->r           = NULL;
+  content->u           = NULL;
+  content->vtemp1      = NULL;
+  content->vtemp2      = NULL;
+  content->vtemp3      = NULL;
+  content->s1          = NULL;
+  content->s2          = NULL;
+  content->ATimes      = NULL;
+  content->ATData      = NULL;
+  content->Psetup      = NULL;
+  content->Psolve      = NULL;
+  content->PData       = NULL;
+  content->print_level = 0;
+  content->info_file   = stdout;
 
-  /* Attach content and ops */
-  S->content = content;
-  S->ops     = ops;
+  /* Allocate content */
+  content->r_star = N_VClone(y);
+  if (content->r_star == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->q = N_VClone(y);
+  if (content->q == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->d = N_VClone(y);
+  if (content->d == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->v = N_VClone(y);
+  if (content->v == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->p = N_VClone(y);
+  if (content->p == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->r = N_VCloneVectorArray(2, y);
+  if (content->r == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->u = N_VClone(y);
+  if (content->u == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->vtemp1 = N_VClone(y);
+  if (content->vtemp1 == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->vtemp2 = N_VClone(y);
+  if (content->vtemp2 == NULL) { SUNLinSolFree(S); return(NULL); }
+
+  content->vtemp3 = N_VClone(y);
+  if (content->vtemp3 == NULL) { SUNLinSolFree(S); return(NULL); }
 
   return(S);
 }
 
 
 /* ----------------------------------------------------------------------------
- * Function to set the type of preconditioning for SPTFQMR to use 
+ * Function to set the type of preconditioning for SPTFQMR to use
  */
 
-SUNDIALS_EXPORT int SUNLinSol_SPTFQMRSetPrecType(SUNLinearSolver S, int pretype) 
+int SUNLinSol_SPTFQMRSetPrecType(SUNLinearSolver S, int pretype)
 {
-  /* Check for legal pretype */ 
-  if ((pretype != PREC_NONE)  && (pretype != PREC_LEFT) &&
-      (pretype != PREC_RIGHT) && (pretype != PREC_BOTH)) {
+  /* Check for legal pretype */
+  if ((pretype != SUN_PREC_NONE)  && (pretype != SUN_PREC_LEFT) &&
+      (pretype != SUN_PREC_RIGHT) && (pretype != SUN_PREC_BOTH)) {
     return(SUNLS_ILL_INPUT);
   }
 
@@ -177,15 +179,15 @@ SUNDIALS_EXPORT int SUNLinSol_SPTFQMRSetPrecType(SUNLinearSolver S, int pretype)
 
 
 /* ----------------------------------------------------------------------------
- * Function to set the maximum number of iterations for SPTFQMR to use 
+ * Function to set the maximum number of iterations for SPTFQMR to use
  */
 
-SUNDIALS_EXPORT int SUNLinSol_SPTFQMRSetMaxl(SUNLinearSolver S, int maxl) 
+int SUNLinSol_SPTFQMRSetMaxl(SUNLinearSolver S, int maxl)
 {
   /* Check for non-NULL SUNLinearSolver */
   if (S == NULL) return(SUNLS_MEM_NULL);
 
-  /* Check for legal pretype */ 
+  /* Check for legal pretype */
   if (maxl <= 0)
     maxl = SUNSPTFQMR_MAXL_DEFAULT;
 
@@ -207,32 +209,49 @@ SUNLinearSolver_Type SUNLinSolGetType_SPTFQMR(SUNLinearSolver S)
 }
 
 
+SUNLinearSolver_ID SUNLinSolGetID_SPTFQMR(SUNLinearSolver S)
+{
+  return(SUNLINEARSOLVER_SPTFQMR);
+}
+
+
 int SUNLinSolInitialize_SPTFQMR(SUNLinearSolver S)
 {
   SUNLinearSolverContent_SPTFQMR content;
 
   /* set shortcut to SPTFQMR memory structure */
-  if (S == NULL) return(SUNLS_MEM_NULL);  
+  if (S == NULL) return(SUNLS_MEM_NULL);
   content = SPTFQMR_CONTENT(S);
 
   /* ensure valid options */
-  if ( (content->pretype != PREC_LEFT) && 
-       (content->pretype != PREC_RIGHT) && 
-       (content->pretype != PREC_BOTH) )
-    content->pretype = PREC_NONE;
-  if (content->maxl <= 0) 
+  if (content->maxl <= 0)
     content->maxl = SUNSPTFQMR_MAXL_DEFAULT;
 
+  if (content->ATimes == NULL) {
+    LASTFLAG(S) = SUNLS_ATIMES_NULL;
+    return(LASTFLAG(S));
+  }
+
+  if ( (content->pretype != SUN_PREC_LEFT) &&
+       (content->pretype != SUN_PREC_RIGHT) &&
+       (content->pretype != SUN_PREC_BOTH) )
+    content->pretype = SUN_PREC_NONE;
+
+  if ((content->pretype != SUN_PREC_NONE) && (content->Psolve == NULL)) {
+    LASTFLAG(S) = SUNLS_PSOLVE_NULL;
+    return(LASTFLAG(S));
+  }
+
   /* no additional memory to allocate */
-  
+
   /* return with success */
   content->last_flag = SUNLS_SUCCESS;
   return(SUNLS_SUCCESS);
 }
 
 
-int SUNLinSolSetATimes_SPTFQMR(SUNLinearSolver S, void* ATData, 
-                               ATimesFn ATimes)
+int SUNLinSolSetATimes_SPTFQMR(SUNLinearSolver S, void* ATData,
+                               SUNATimesFn ATimes)
 {
   /* set function pointers to integrator-supplied ATimes routine
      and data, and return with success */
@@ -245,7 +264,7 @@ int SUNLinSolSetATimes_SPTFQMR(SUNLinearSolver S, void* ATData,
 
 
 int SUNLinSolSetPreconditioner_SPTFQMR(SUNLinearSolver S, void* PData,
-                                       PSetupFn Psetup, PSolveFn Psolve)
+                                       SUNPSetupFn Psetup, SUNPSolveFn Psolve)
 {
   /* set function pointers to integrator-supplied Psetup and PSolve
      routines and data, and return with success */
@@ -262,7 +281,7 @@ int SUNLinSolSetScalingVectors_SPTFQMR(SUNLinearSolver S,
                                        N_Vector s1,
                                        N_Vector s2)
 {
-  /* set N_Vector pointers to integrator-supplied scaling vectors, 
+  /* set N_Vector pointers to integrator-supplied scaling vectors,
      and return with success */
   if (S == NULL) return(SUNLS_MEM_NULL);
   SPTFQMR_CONTENT(S)->s1 = s1;
@@ -272,34 +291,44 @@ int SUNLinSolSetScalingVectors_SPTFQMR(SUNLinearSolver S,
 }
 
 
+int SUNLinSolSetZeroGuess_SPTFQMR(SUNLinearSolver S, booleantype onoff)
+{
+  /* set flag indicating a zero initial guess */
+  if (S == NULL) return(SUNLS_MEM_NULL);
+  SPTFQMR_CONTENT(S)->zeroguess = onoff;
+  LASTFLAG(S) = SUNLS_SUCCESS;
+  return(LASTFLAG(S));
+}
+
+
 int SUNLinSolSetup_SPTFQMR(SUNLinearSolver S, SUNMatrix A)
 {
   int ier;
-  PSetupFn Psetup;
+  SUNPSetupFn Psetup;
   void* PData;
 
   /* Set shortcuts to SPTFQMR memory structures */
   if (S == NULL) return(SUNLS_MEM_NULL);
   Psetup = SPTFQMR_CONTENT(S)->Psetup;
   PData = SPTFQMR_CONTENT(S)->PData;
-  
-  /* no solver-specific setup is required, but if user-supplied 
+
+  /* no solver-specific setup is required, but if user-supplied
      Psetup routine exists, call that here */
   if (Psetup != NULL) {
     ier = Psetup(PData);
     if (ier != 0) {
-      LASTFLAG(S) = (ier < 0) ? 
-	SUNLS_PSET_FAIL_UNREC : SUNLS_PSET_FAIL_REC;
+      LASTFLAG(S) = (ier < 0) ?
+        SUNLS_PSET_FAIL_UNREC : SUNLS_PSET_FAIL_REC;
       return(LASTFLAG(S));
     }
   }
-  
-  /* return with success */ 
+
+  /* return with success */
   return(SUNLS_SUCCESS);
 }
 
 
-int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x, 
+int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
                            N_Vector b, realtype delta)
 {
   /* local data and shortcut variables */
@@ -307,12 +336,12 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   realtype rho[2];
   realtype r_init_norm, r_curr_norm;
   realtype temp_val;
-  booleantype preOnLeft, preOnRight, scale_x, scale_b, converged;
-  booleantype b_ok;
+  booleantype preOnLeft, preOnRight, scale_x, scale_b, converged, b_ok;
+  booleantype *zeroguess;
   int n, m, ier, l_max;
   void *A_data, *P_data;
-  ATimesFn atimes;
-  PSolveFn psolve;
+  SUNATimesFn atimes;
+  SUNPSolveFn psolve;
   realtype *res_norm;
   int *nli;
   N_Vector sx, sb, r_star, q, d, v, p, *r, u, vtemp1, vtemp2, vtemp3;
@@ -320,7 +349,7 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   /* local variables for fused vector operations */
   realtype cv[3];
   N_Vector Xv[3];
-  
+
   /* Make local shorcuts to solver variables. */
   if (S == NULL) return(SUNLS_MEM_NULL);
   l_max        = SPTFQMR_CONTENT(S)->maxl;
@@ -340,6 +369,7 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   P_data       = SPTFQMR_CONTENT(S)->PData;
   atimes       = SPTFQMR_CONTENT(S)->ATimes;
   psolve       = SPTFQMR_CONTENT(S)->Psolve;
+  zeroguess    = &(SPTFQMR_CONTENT(S)->zeroguess);
   nli          = &(SPTFQMR_CONTENT(S)->numiters);
   res_norm     = &(SPTFQMR_CONTENT(S)->resnorm);
 
@@ -350,19 +380,47 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   b_ok = SUNFALSE;
 
   /* set booleantype flags for internal solver options */
-  preOnLeft  = ( (SPTFQMR_CONTENT(S)->pretype == PREC_LEFT) || 
-                 (SPTFQMR_CONTENT(S)->pretype == PREC_BOTH) );
-  preOnRight = ( (SPTFQMR_CONTENT(S)->pretype == PREC_RIGHT) || 
-                 (SPTFQMR_CONTENT(S)->pretype == PREC_BOTH) );
+  preOnLeft  = ( (SPTFQMR_CONTENT(S)->pretype == SUN_PREC_LEFT) ||
+                 (SPTFQMR_CONTENT(S)->pretype == SUN_PREC_BOTH) );
+  preOnRight = ( (SPTFQMR_CONTENT(S)->pretype == SUN_PREC_RIGHT) ||
+                 (SPTFQMR_CONTENT(S)->pretype == SUN_PREC_BOTH) );
   scale_x = (sx != NULL);
   scale_b = (sb != NULL);
 
+  /* Check for unsupported use case */
+  if (preOnRight && !(*zeroguess)) {
+    *zeroguess  = SUNFALSE;
+    LASTFLAG(S) = SUNLS_ILL_INPUT;
+    return(SUNLS_ILL_INPUT);
+  }
+
+#ifdef SUNDIALS_BUILD_WITH_MONITORING
+  if (SPTFQMR_CONTENT(S)->print_level && SPTFQMR_CONTENT(S)->info_file)
+    STAN_SUNDIALS_FPRINTF(SPTFQMR_CONTENT(S)->info_file, "SUNLINSOL_SPTFQMR:\n");
+#endif
+
+  /* Check if Atimes function has been set */
+  if (atimes == NULL) {
+    *zeroguess  = SUNFALSE;
+    LASTFLAG(S) = SUNLS_ATIMES_NULL;
+    return(LASTFLAG(S));
+  }
+
+  /* If preconditioning, check if psolve has been set */
+  if ((preOnLeft || preOnRight) && psolve == NULL) {
+    *zeroguess  = SUNFALSE;
+    LASTFLAG(S) = SUNLS_PSOLVE_NULL;
+    return(LASTFLAG(S));
+  }
+
   /* Set r_star to initial (unscaled) residual r_star = r_0 = b - A*x_0 */
   /* NOTE: if x == 0 then just set residual to b and continue */
-  if (N_VDotProd(x, x) == ZERO) N_VScale(ONE, b, r_star);
-  else {
+  if (*zeroguess) {
+    N_VScale(ONE, b, r_star);
+  } else {
     ier = atimes(A_data, x, r_star);
     if (ier != 0) {
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = (ier < 0) ?
         SUNLS_ATIMES_FAIL_UNREC : SUNLS_ATIMES_FAIL_REC;
       return(LASTFLAG(S));
@@ -372,8 +430,9 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
 
   /* Apply left preconditioner and b-scaling to r_star (or really just r_0) */
   if (preOnLeft) {
-    ier = psolve(P_data, r_star, vtemp1, delta, PREC_LEFT);
+    ier = psolve(P_data, r_star, vtemp1, delta, SUN_PREC_LEFT);
     if (ier != 0) {
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = (ier < 0) ?
         SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
       return(LASTFLAG(S));
@@ -392,7 +451,19 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   /* Compute norm of initial residual (r_0) to see if we really need
      to do anything */
   *res_norm = r_init_norm = SUNRsqrt(rho[0]);
+
+#ifdef SUNDIALS_BUILD_WITH_MONITORING
+  /* print initial residual */
+  if (SPTFQMR_CONTENT(S)->print_level && SPTFQMR_CONTENT(S)->info_file)
+  {
+    STAN_SUNDIALS_FPRINTF(SPTFQMR_CONTENT(S)->info_file,
+            SUNLS_MSG_RESIDUAL,
+            (long int) 0, *res_norm);
+  }
+#endif
+
   if (r_init_norm <= delta) {
+    *zeroguess  = SUNFALSE;
     LASTFLAG(S) = SUNLS_SUCCESS;
     return(LASTFLAG(S));
   }
@@ -402,8 +473,9 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   else N_VScale(ONE, r_star, vtemp1);
   if (preOnRight) {
     N_VScale(ONE, vtemp1, v);
-    ier = psolve(P_data, v, vtemp1, delta, PREC_RIGHT);
+    ier = psolve(P_data, v, vtemp1, delta, SUN_PREC_RIGHT);
     if (ier != 0) {
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = (ier < 0) ?
         SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
       return(LASTFLAG(S));
@@ -411,13 +483,15 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   }
   ier = atimes(A_data, vtemp1, v);
   if (ier != 0) {
+    *zeroguess  = SUNFALSE;
     LASTFLAG(S) = (ier < 0) ?
       SUNLS_ATIMES_FAIL_UNREC : SUNLS_ATIMES_FAIL_REC;
     return(LASTFLAG(S));
   }
   if (preOnLeft) {
-    ier = psolve(P_data, v, vtemp1, delta, PREC_LEFT);
+    ier = psolve(P_data, v, vtemp1, delta, SUN_PREC_LEFT);
     if (ier != 0) {
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = (ier < 0) ?
         SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
       return(LASTFLAG(S));
@@ -433,6 +507,9 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   N_VScale(ONE, r_star, p);
   N_VConst(ZERO, d);
 
+  /* Set x = sx x if non-zero guess */
+  if (scale_x && !(*zeroguess)) N_VProd(sx, x, x);
+
   tau = r_init_norm;
   v_bar = eta = ZERO;
 
@@ -456,8 +533,9 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
     if (scale_x) N_VDiv(r[1], sx, r[1]);
     if (preOnRight) {
       N_VScale(ONE, r[1], vtemp1);
-      ier = psolve(P_data, vtemp1, r[1], delta, PREC_RIGHT);
+      ier = psolve(P_data, vtemp1, r[1], delta, SUN_PREC_RIGHT);
       if (ier != 0) {
+        *zeroguess  = SUNFALSE;
         LASTFLAG(S) = (ier < 0) ?
           SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
         return(LASTFLAG(S));
@@ -465,13 +543,15 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
     }
     ier = atimes(A_data, r[1], vtemp1);
     if (ier != 0) {
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = (ier < 0) ?
         SUNLS_ATIMES_FAIL_UNREC : SUNLS_ATIMES_FAIL_REC;
       return(LASTFLAG(S));
     }
     if (preOnLeft) {
-      ier = psolve(P_data, vtemp1, r[1], delta, PREC_LEFT);
+      ier = psolve(P_data, vtemp1, r[1], delta, SUN_PREC_LEFT);
       if (ier != 0) {
+        *zeroguess  = SUNFALSE;
         LASTFLAG(S) = (ier < 0) ?
           SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
         return(LASTFLAG(S));
@@ -492,13 +572,13 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
        *       if the inner loop is executed twice
        */
       if (m == 0) {
-	temp_val = SUNRsqrt(N_VDotProd(r[1], r[1]));
-	omega = SUNRsqrt(SUNRsqrt(N_VDotProd(r[0], r[0]))*temp_val);
-	N_VLinearSum(ONE, u, SUNSQR(v_bar)*eta/alpha, d, d);
+        temp_val = SUNRsqrt(N_VDotProd(r[1], r[1]));
+        omega = SUNRsqrt(SUNRsqrt(N_VDotProd(r[0], r[0]))*temp_val);
+        N_VLinearSum(ONE, u, SUNSQR(v_bar)*eta/alpha, d, d);
       }
       else {
-	omega = temp_val;
-	N_VLinearSum(ONE, q, SUNSQR(v_bar)*eta/alpha, d, d);
+        omega = temp_val;
+        N_VLinearSum(ONE, q, SUNSQR(v_bar)*eta/alpha, d, d);
       }
 
       /* v_bar = omega/tau */
@@ -514,17 +594,30 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
       eta = SUNSQR(c)*alpha;
 
       /* x = x+eta*d */
-      N_VLinearSum(ONE, x, eta, d, x);
+      if (n == 0 && m == 0 && *zeroguess)
+        N_VScale(eta, d, x);
+      else
+        N_VLinearSum(ONE, x, eta, d, x);
 
       /* Check for convergence... */
       /* NOTE: just use approximation to norm of residual, if possible */
       *res_norm = r_curr_norm = tau*SUNRsqrt(m+1);
 
+#ifdef SUNDIALS_BUILD_WITH_MONITORING
+      /* print current iteration number and the residual */
+      if (SPTFQMR_CONTENT(S)->print_level && SPTFQMR_CONTENT(S)->info_file)
+      {
+        STAN_SUNDIALS_FPRINTF(SPTFQMR_CONTENT(S)->info_file,
+                SUNLS_MSG_RESIDUAL,
+                (long int) *nli, *res_norm);
+      }
+#endif
+
       /* Exit inner loop if iteration has converged based upon approximation
-	 to norm of current residual */
+         to norm of current residual */
       if (r_curr_norm <= delta) {
-	converged = SUNTRUE;
-	break;
+        converged = SUNTRUE;
+        break;
       }
 
       /* Decide if actual norm of residual vector should be computed */
@@ -540,60 +633,64 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
        *       number of psovles if using left preconditioning
        */
       if ((r_curr_norm > delta) ||
-	  (r_curr_norm >= r_init_norm && m == 1 && n == l_max)) {
+          (r_curr_norm >= r_init_norm && m == 1 && n == l_max)) {
 
-	/* Compute norm of residual ||b-A*x||_2 (preconditioned and scaled) */
-	if (scale_x) N_VDiv(x, sx, vtemp1);
-	else N_VScale(ONE, x, vtemp1);
-	if (preOnRight) {
-	  ier = psolve(P_data, vtemp1, vtemp2, delta, PREC_RIGHT);
-	  if (ier != 0) {
+        /* Compute norm of residual ||b-A*x||_2 (preconditioned and scaled) */
+        if (scale_x) N_VDiv(x, sx, vtemp1);
+        else N_VScale(ONE, x, vtemp1);
+        if (preOnRight) {
+          ier = psolve(P_data, vtemp1, vtemp2, delta, SUN_PREC_RIGHT);
+          if (ier != 0) {
+            *zeroguess  = SUNFALSE;
             LASTFLAG(S) = (ier < 0) ?
               SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_UNREC;
             return(LASTFLAG(S));
           }
-	  N_VScale(ONE, vtemp2, vtemp1);
-	}
-	ier = atimes(A_data, vtemp1, vtemp2);
+          N_VScale(ONE, vtemp2, vtemp1);
+        }
+        ier = atimes(A_data, vtemp1, vtemp2);
         if (ier != 0) {
+          *zeroguess  = SUNFALSE;
           LASTFLAG(S) = (ier < 0) ?
             SUNLS_ATIMES_FAIL_UNREC : SUNLS_ATIMES_FAIL_REC;
           return(LASTFLAG(S));
         }
-	if (preOnLeft) {
-	  ier = psolve(P_data, vtemp2, vtemp1, delta, PREC_LEFT);
-	  if (ier != 0) {
+        if (preOnLeft) {
+          ier = psolve(P_data, vtemp2, vtemp1, delta, SUN_PREC_LEFT);
+          if (ier != 0) {
+            *zeroguess  = SUNFALSE;
             LASTFLAG(S) = (ier < 0) ?
               SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
             return(LASTFLAG(S));
           }
-	}
-	else N_VScale(ONE, vtemp2, vtemp1);
-	if (scale_b) N_VProd(sb, vtemp1, vtemp2);
-	else N_VScale(ONE, vtemp1, vtemp2);
-	/* Only precondition and scale b once (result saved for reuse) */
-	if (!b_ok) {
-	  b_ok = SUNTRUE;
-	  if (preOnLeft) {
-	    ier = psolve(P_data, b, vtemp3, delta, PREC_LEFT);
-	    if (ier != 0) {
+        }
+        else N_VScale(ONE, vtemp2, vtemp1);
+        if (scale_b) N_VProd(sb, vtemp1, vtemp2);
+        else N_VScale(ONE, vtemp1, vtemp2);
+        /* Only precondition and scale b once (result saved for reuse) */
+        if (!b_ok) {
+          b_ok = SUNTRUE;
+          if (preOnLeft) {
+            ier = psolve(P_data, b, vtemp3, delta, SUN_PREC_LEFT);
+            if (ier != 0) {
+              *zeroguess  = SUNFALSE;
               LASTFLAG(S) = (ier < 0) ?
                 SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
               return(LASTFLAG(S));
             }
-	  }
-	  else N_VScale(ONE, b, vtemp3);
-	  if (scale_b) N_VProd(sb, vtemp3, vtemp3);
-	}
-	N_VLinearSum(ONE, vtemp3, -ONE, vtemp2, vtemp1);
-	*res_norm = r_curr_norm = SUNRsqrt(N_VDotProd(vtemp1, vtemp1));
+          }
+          else N_VScale(ONE, b, vtemp3);
+          if (scale_b) N_VProd(sb, vtemp3, vtemp3);
+        }
+        N_VLinearSum(ONE, vtemp3, -ONE, vtemp2, vtemp1);
+        *res_norm = r_curr_norm = SUNRsqrt(N_VDotProd(vtemp1, vtemp1));
 
-	/* Exit inner loop if inequality condition is satisfied 
-	   (meaning exit if we have converged) */
-	if (r_curr_norm <= delta) {
-	  converged = SUNTRUE;
-	  break;
-	}
+        /* Exit inner loop if inequality condition is satisfied
+           (meaning exit if we have converged) */
+        if (r_curr_norm <= delta) {
+          converged = SUNTRUE;
+          break;
+        }
 
       }
 
@@ -622,15 +719,20 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
     Xv[2] = u;
 
     ier = N_VLinearCombination(3, cv, Xv, p);
-    if (ier != SUNLS_SUCCESS) return(SUNLS_VECTOROP_ERR);
+    if (ier != SUNLS_SUCCESS) {
+      *zeroguess  = SUNFALSE;
+      LASTFLAG(S) = SUNLS_VECTOROP_ERR;
+      return(SUNLS_VECTOROP_ERR);
+    }
 
     /* v = A*p */
     if (scale_x) N_VDiv(p, sx, vtemp1);
     else N_VScale(ONE, p, vtemp1);
     if (preOnRight) {
       N_VScale(ONE, vtemp1, v);
-      ier = psolve(P_data, v, vtemp1, delta, PREC_RIGHT);
+      ier = psolve(P_data, v, vtemp1, delta, SUN_PREC_RIGHT);
       if (ier != 0) {
+        *zeroguess  = SUNFALSE;
         LASTFLAG(S) = (ier < 0) ?
           SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
         return(LASTFLAG(S));
@@ -638,13 +740,15 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
     }
     ier = atimes(A_data, vtemp1, v);
     if (ier != 0) {
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = (ier < 0) ?
         SUNLS_ATIMES_FAIL_UNREC : SUNLS_ATIMES_FAIL_REC;
       return(LASTFLAG(S));
     }
     if (preOnLeft) {
-      ier = psolve(P_data, v, vtemp1, delta, PREC_LEFT);
+      ier = psolve(P_data, v, vtemp1, delta, SUN_PREC_LEFT);
       if (ier != 0) {
+        *zeroguess  = SUNFALSE;
         LASTFLAG(S) = (ier < 0) ?
           SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
         return(LASTFLAG(S));
@@ -666,22 +770,26 @@ int SUNLinSolSolve_SPTFQMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   if ((converged == SUNTRUE) || (r_curr_norm < r_init_norm)) {
     if (scale_x) N_VDiv(x, sx, x);
     if (preOnRight) {
-      ier = psolve(P_data, x, vtemp1, delta, PREC_RIGHT);
+      ier = psolve(P_data, x, vtemp1, delta, SUN_PREC_RIGHT);
       if (ier != 0) {
+        *zeroguess  = SUNFALSE;
         LASTFLAG(S) = (ier < 0) ?
           SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_UNREC;
         return(LASTFLAG(S));
       }
       N_VScale(ONE, vtemp1, x);
     }
-    if (converged == SUNTRUE) 
+
+    *zeroguess = SUNFALSE;
+    if (converged == SUNTRUE)
       LASTFLAG(S) = SUNLS_SUCCESS;
-    else 
+    else
       LASTFLAG(S) = SUNLS_RES_REDUCED;
     return(LASTFLAG(S));
   }
   /* Otherwise, return error code */
   else {
+    *zeroguess  = SUNFALSE;
     LASTFLAG(S) = SUNLS_CONV_FAIL;
     return(LASTFLAG(S));
   }
@@ -711,7 +819,7 @@ N_Vector SUNLinSolResid_SPTFQMR(SUNLinearSolver S)
 }
 
 
-long int SUNLinSolLastFlag_SPTFQMR(SUNLinearSolver S)
+sunindextype SUNLinSolLastFlag_SPTFQMR(SUNLinearSolver S)
 {
   /* return the stored 'last_flag' value */
   if (S == NULL) return(-1);
@@ -719,8 +827,8 @@ long int SUNLinSolLastFlag_SPTFQMR(SUNLinearSolver S)
 }
 
 
-int SUNLinSolSpace_SPTFQMR(SUNLinearSolver S, 
-                           long int *lenrwLS, 
+int SUNLinSolSpace_SPTFQMR(SUNLinearSolver S,
+                           long int *lenrwLS,
                            long int *leniwLS)
 {
   sunindextype liw1, lrw1;
@@ -737,31 +845,91 @@ int SUNLinSolFree_SPTFQMR(SUNLinearSolver S)
 {
   if (S == NULL) return(SUNLS_SUCCESS);
 
-  /* delete items from within the content structure */
-  if (SPTFQMR_CONTENT(S)->r_star)
-    N_VDestroy(SPTFQMR_CONTENT(S)->r_star);
-  if (SPTFQMR_CONTENT(S)->q)
-    N_VDestroy(SPTFQMR_CONTENT(S)->q);
-  if (SPTFQMR_CONTENT(S)->d)
-    N_VDestroy(SPTFQMR_CONTENT(S)->d);
-  if (SPTFQMR_CONTENT(S)->v)
-    N_VDestroy(SPTFQMR_CONTENT(S)->v);
-  if (SPTFQMR_CONTENT(S)->p)
-    N_VDestroy(SPTFQMR_CONTENT(S)->p);
-  if (SPTFQMR_CONTENT(S)->r)
-    N_VDestroyVectorArray(SPTFQMR_CONTENT(S)->r, 2);
-  if (SPTFQMR_CONTENT(S)->u)
-    N_VDestroy(SPTFQMR_CONTENT(S)->u);
-  if (SPTFQMR_CONTENT(S)->vtemp1)
-    N_VDestroy(SPTFQMR_CONTENT(S)->vtemp1);
-  if (SPTFQMR_CONTENT(S)->vtemp2)
-    N_VDestroy(SPTFQMR_CONTENT(S)->vtemp2);
-  if (SPTFQMR_CONTENT(S)->vtemp3)
-    N_VDestroy(SPTFQMR_CONTENT(S)->vtemp3);
-
-  /* delete generic structures */
-  free(S->content);  S->content = NULL;
-  free(S->ops);  S->ops = NULL;
+  if (S->content) {
+    /* delete items from within the content structure */
+    if (SPTFQMR_CONTENT(S)->r_star) {
+      N_VDestroy(SPTFQMR_CONTENT(S)->r_star);
+      SPTFQMR_CONTENT(S)->r_star = NULL;
+    }
+    if (SPTFQMR_CONTENT(S)->q) {
+      N_VDestroy(SPTFQMR_CONTENT(S)->q);
+      SPTFQMR_CONTENT(S)->q = NULL;
+    }
+    if (SPTFQMR_CONTENT(S)->d) {
+      N_VDestroy(SPTFQMR_CONTENT(S)->d);
+      SPTFQMR_CONTENT(S)->d = NULL;
+    }
+    if (SPTFQMR_CONTENT(S)->v) {
+      N_VDestroy(SPTFQMR_CONTENT(S)->v);
+      SPTFQMR_CONTENT(S)->v = NULL;
+    }
+    if (SPTFQMR_CONTENT(S)->p) {
+      N_VDestroy(SPTFQMR_CONTENT(S)->p);
+      SPTFQMR_CONTENT(S)->p = NULL;
+    }
+    if (SPTFQMR_CONTENT(S)->r) {
+      N_VDestroyVectorArray(SPTFQMR_CONTENT(S)->r, 2);
+      SPTFQMR_CONTENT(S)->r = NULL;
+    }
+    if (SPTFQMR_CONTENT(S)->u) {
+      N_VDestroy(SPTFQMR_CONTENT(S)->u);
+      SPTFQMR_CONTENT(S)->u = NULL;
+    }
+    if (SPTFQMR_CONTENT(S)->vtemp1) {
+      N_VDestroy(SPTFQMR_CONTENT(S)->vtemp1);
+      SPTFQMR_CONTENT(S)->vtemp1 = NULL;
+    }
+    if (SPTFQMR_CONTENT(S)->vtemp2) {
+      N_VDestroy(SPTFQMR_CONTENT(S)->vtemp2);
+      SPTFQMR_CONTENT(S)->vtemp2 = NULL;
+    }
+    if (SPTFQMR_CONTENT(S)->vtemp3) {
+      N_VDestroy(SPTFQMR_CONTENT(S)->vtemp3);
+      SPTFQMR_CONTENT(S)->vtemp3 = NULL;
+    }
+    free(S->content); S->content = NULL;
+  }
+  if (S->ops) { free(S->ops); S->ops = NULL; }
   free(S); S = NULL;
   return(SUNLS_SUCCESS);
 }
+
+
+int SUNLinSolSetInfoFile_SPTFQMR(SUNLinearSolver S,
+                                 FILE* info_file)
+{
+#ifdef SUNDIALS_BUILD_WITH_MONITORING
+  /* check that the linear solver is non-null */
+  if (S == NULL)
+    return(SUNLS_MEM_NULL);
+
+  SPTFQMR_CONTENT(S)->info_file = info_file;
+
+  return(SUNLS_SUCCESS);
+#else
+  SUNDIALS_DEBUG_PRINT("ERROR in SUNLinSolSetInfoFile_SPTFQMR: SUNDIALS was not built with monitoring\n");
+  return(SUNLS_ILL_INPUT);
+#endif
+}
+
+
+int SUNLinSolSetPrintLevel_SPTFQMR(SUNLinearSolver S,
+                                   int print_level)
+{
+#ifdef SUNDIALS_BUILD_WITH_MONITORING
+  /* check that the linear solver is non-null */
+  if (S == NULL)
+    return(SUNLS_MEM_NULL);
+
+  /* check for valid print level */
+  if (print_level < 0 || print_level > 1)
+    return(SUNLS_ILL_INPUT);
+
+  SPTFQMR_CONTENT(S)->print_level = print_level;
+
+  return(SUNLS_SUCCESS);
+#else
+  SUNDIALS_DEBUG_PRINT("ERROR in SUNLinSolSetPrintLevel_SPTFQMR: SUNDIALS was not built with monitoring\n");
+  return(SUNLS_ILL_INPUT);
+#endif
+}
